/*
 *	analy.cpp
 */

/*	Copyright (C) 1999 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "htstring.h"
#include "global.h"
#include "fs.h"
#include "htunasm.h"
#include "codeanaly.h"
#include "dataanaly.h"
#include "analy.h"

int some_analyser_active = 0;
/*
 *
 */
void	analyser::init()
{
	addr_queue = new tqueue();
	addr_queue->init();
	addrs = NULL;
	ops_parsed = 0;
	next_explored = first_explored = last_explored = 0xffffffff;
	explored = new	tarea();
	explored->init();
	active = false;
     init_codeanalyser();
     init_dataanalyser();
     init_unasm();
};


void loadaddrs(htfile *f, taddr **addr, int level, int *left)
{
    	if ((*left)<=0) {
          *addr = NULL;
     	return;
     }

   	*addr = (taddr *) calloc(sizeof(taddr), 1);


	if ((level > 1) && (*left > 1)) loadaddrs(f, &(*addr)->left, level / 2, left);

     f->read(&(*addr)->addr, sizeof(ADDR));
     fstrget(f, &(*addr)->label);
     // xrefs
     (*addr)->xreflist = NULL;
     txref	**xref = &(*addr)->xreflist;
     int		count = 0;
     f->read(&count, sizeof(int));
     while (count--) {
       	f->read(&(*xref)->addr, sizeof(ADDR));
          f->read(&(*xref)->xreftype, sizeof(txreftype));
          xref = &(*xref)->next;
     }
     xref = NULL;
     // comments
     (*addr)->comments = NULL;
     tcomment	**com = &(*addr)->comments;
     count = 0;
     f->read(&count, sizeof(int));
     while (count--) {
     	fstrget(f, &(*com)->text);
          com = &(*com)->next;
     }
     com = NULL;

     f->read(&(*addr)->type, sizeof(taddrtype));

     (*left)--;

	if ((level > 1) && (*left > 1)) loadaddrs(f, &(*addr)->right, level / 2, left);
}
/*
 *
 */
void	analyser::load(htfile *f)
{
     addr_queue->load(f);
     f->read(&ops_parsed, sizeof(int));
     f->read(&active, sizeof(bool));
     f->read(&next_explored, sizeof(ADDR));
     f->read(&first_explored, sizeof(ADDR));
     f->read(&last_explored, sizeof(ADDR));
     explored->load(f);
     int count;
     f->read(&count, sizeof(int));
	loadaddrs(f, &addrs, count, &count);
	code->load(f);
	data->load(f);
     unasm->load(f);
};

/*
 *
 */
void	addrsdone(taddr *addrs)
{
	if (addrs) {
		addrsdone(addrs->left);
		addrsdone(addrs->right);
		free(addrs->label);
		free(addrs);
	}
};

/*
 *
 */
void	analyser::done()
{
	addrsdone(addrs);
	addr_queue->done();
	delete addr_queue;
	explored->done();
	delete explored;
     code->done();
     delete code;
     data->done();
     delete data;
};

/*
 *	true meens new label created
 */
bool	analyser::addaddrlabel(ADDR Addr, char *prefix)
{
	if (!validaddr(Addr, scvalid)) return false;

	char	*label = (char	*)malloc(9+strlen(prefix));
	char	num[9];

	sprintf(num, HEX8FORMAT, Addr);
	strcpy(label, prefix);
	strcat(label, num);

	char *result;

	return	addlabel(Addr, label, &result);
}

/*
 *
 */
void	analyser::addcomment(ADDR Addr, int line, char *c)
{
	// line 0 meens append (atm assume append every time ;-))

	if (!validaddr(Addr, scvalid)) return;

	taddr	*a =	newaddr(Addr);

	tcomment *com = a->comments;

	if (com) {
		while (com->next) com = com->next;
		com->next	= (tcomment *)	malloc(sizeof(tcomment));
		com = com->next;
	} else {
		com = (tcomment *) malloc(sizeof(tcomment));
		a->comments = com;
	};

	com->next	= NULL;
	com->text	= c;

	dprintf("#(%08lx) comment `%s'\n", Addr, c);
};

/*
 *  true  - new label has been created
 *  false - label has been overwritten -- doesnt work yet
 *
 */
bool	analyser::addlabel(ADDR Addr,	char	*label)
{
	if (!validaddr(Addr, scvalid)) return false;

	taddr *a = newaddr(Addr);

     if (a->label) free(a->label);

	a->label = label;

	dprintf("label `%s' assigned to %08lx\n",	label, Addr);

     return true;
};

/*
 *  true  - new label has been created
 *  false - returned old label
 *
 */
bool	analyser::addlabel(ADDR Addr,	char	*label, char **result)
{
	if (!validaddr(Addr, scvalid)) return false;

	taddr *a = newaddr(Addr);

     if (a->label) {
     	*result = a->label;
          return false;
     }

	a->label = label;

	dprintf("label `%s' assigned to %08lx\n",	label, Addr);
	return true;
};

/*
 *
 */
void	analyser::addxref(ADDR from, ADDR to, txreftype action)
{
	if ((!validaddr(from, scvalid)) ||	(!validaddr(to, scvalid))) return;

	taddr	*a =	newaddr(from);

	txref	*x = a->xreflist;

	if (x) {
		while (x->next) x =	x->next;
		x->next =	(txref *)	malloc(sizeof(txref));
		x = x->next;
	} else {
		x = (txref *) malloc(sizeof(txref));
		a->xreflist = x;
	};

	x->next =	NULL;
	x->addr =	to;
	x->xreftype = action;

	dprintf("xref %08lx->%08lx\n", from, to);
};

/*
 *
 */
void	analyser::beginanalysis(ADDR Addr)
{
	dprintf("################################\nAnalysis started at %08lx:\n", Addr);
	first_explored	= last_explored = Addr;

     dprintf("KLOPOK: %d\n", validaddr(0x00401400, scvalid));

	if (gotoaddr(Addr)) setactive(true);
};

/*
 *
 */
ADDR	analyser::branchaddr(char *opcode)
{
	// abstract
     return 0;
};

/*
 *
 */
void	*analyser::bufptr(ADDR Addr)
{
	// abstract
     return 0;
};

/*
 *
 */
bool	analyser::continueanalysis()
{
	void			*buf;
	char			*instr;
	int			len;
	tbranchtype	branch;

	if (!active) return	false;
	do {
		if (!savetychangeaddr(addr)) {
          	setactive(false);
               return false;
		}

		buf = bufptr(addr);
		instr = unasm->getstring((byte *)buf, 15, mapaddr(addr), &len);
		dprintf("%08lx  %s\n", addr, instr);
		ops_parsed++;

		last_explored += len;
		branch = isbranch(instr);
		if (branch != brnobranch) dobranch(branch, instr,	len); else {
			examineopcode(instr);
			addr	+= len;
		};
	} while ((ops_parsed % MAX_OPS_PER_CONTINUE) !=0);
	notifyprogress(addr);
	return true;
};

/*
 *
 */
void analyser::continueanalysisat(ADDR Addr)
{
	if (!validcodeaddr(Addr)) return;
     if (active) {
     	pushaddr(Addr);
     } else {
		beginanalysis(Addr);
     }
}

/*
 *
 */
void	analyser::dataaccess(ADDR Addr, taccess	access)
{

	if (!validaddr(Addr, scvalid)) {
/*		char	msg[100];
		sprintf(msg, "access of invalid addr %08lx at addr %08lx", Addr, addr);
		log(msg);*/
		return;
	};

	pushaddr(Addr);                       // test

	addaddrlabel(Addr, "data_");
	dprintf("ACCESS of -%08lx-\n", Addr);
};

/*
 *
 */
void	analyser::dobranch(tbranchtype branch, char *opcode, int len)
{

	ADDR	branch_addr = branchaddr(opcode);

/*	if (branch != brreturn) {
		if (!validcodeaddr(branch_addr)) {
			char	msg[100];
			sprintf(msg, "branch to invalid addr %08lx from addr %08lx", branch_addr, addr);
			log(msg);
		}
	}*/
	switch (branch) {
		case	brjump: {
			addxref(branch_addr, addr, xrefjump);
			if (addaddrlabel(branch_addr, "loc_")) addcomment(branch_addr, 0, " ");
			gotoaddr(branch_addr);
			break;
		}
		case	brreturn:	{
			gotoaddr(INVALID_ADDR);
			break;
		}
		case	brcall: {
		    	addxref(branch_addr, addr, xrefcall);
			if (addaddrlabel(branch_addr,	"sub_")) {
			     addcomment(branch_addr, 0, " ");
			  	addcomment(branch_addr, 0, ";-----------------------");
    				addcomment(branch_addr, 0, ";  S U B R O U T I N E");
	    		  	addcomment(branch_addr, 0, ";-----------------------");
			}
	    	  	pushaddr(addr+len);
	   		gotoaddr(branch_addr);
			break;
		}
		case	brjXX: {
	 			addxref(branch_addr, addr, xrefjump);
			if (addaddrlabel(branch_addr, "loc_")) addcomment(branch_addr, 0, " ");
	    		pushaddr(addr+len);
	    		gotoaddr(branch_addr);
			break;
		}
          default:{}
	}
};

/*
 *
 */
taddrtypetype analyser::examinedata(ADDR addr)
{
	if ((validreadaddr(addr)) && (validaddr(addr, scinitialized))) {


     } else return dtunknown;
     return dtunknown;
};

/*
 *
 */
void	analyser::examineopcode(char *opcode)
{
	/*
      *	very abstract
      *
      *		this function is called for each opcode parsed and has to
      *		call dataaccess for EVERY addr opcode contains.
      *
      */
};

/*
 *
 */
char	*analyser::getaddrlabel(ADDR Addr)
{
	taddr	*a = searchaddr(Addr);

	return (a) ? a->label : NULL;
};

/*
 *
 */
bool	analyser::gotoaddr(ADDR Addr)
{
	if (first_explored < last_explored) {
		dprintf("Explored Add %08lx - %08lx\n", first_explored, last_explored);
		explored->add(first_explored, last_explored);
     }

	addr	= Addr;

	if ((explored->contains(addr)) || (!validcodeaddr(addr))) {
		dprintf("Address: %08lx Valid: %d Explored: %d\n", addr, validcodeaddr(addr), explored->contains(addr));
		while ((explored->contains(addr)) || (!validcodeaddr(addr))){
	 		if (addr_queue->empty())	{
				first_explored	= last_explored = INVALID_ADDR;
				addr	= INVALID_ADDR;
				return false;
			};
			addr	= (ADDR) addr_queue->dequeue();
			dprintf("pop %08lx   (ex: %d  v: %d)\n", addr,	explored->contains(addr), validcodeaddr(addr));
	   	};
	};

	next_explored = explored->findnext(addr);
	first_explored	= last_explored = addr;
	return true;
};

/*
 *
 */
void analyser::init_codeanalyser()
{
	code = new codeanalyser();
     code->init(this);
}

/*
 *
 */
void analyser::init_dataanalyser()
{
     data = new dataanalyser();
     data->init(this);
}

/*
 *
 */
void analyser::init_unasm()
{
	// abstract
}

/*
 *
 */
tbranchtype analyser::isbranch(char *opcode)
{
	// abstract
     return brnobranch;
};

/*
 *
 */
void	analyser::log(char *s)
{
	// stub
};

/*
 *
 */
ADDR	analyser::mapaddr(ADDR Addr)
{
	/*
      * 	this function should map the independent address Addr to a
      * 	processor dependent
      * 	e.g.    .23423 --> 00234324    (or something like this)
      *   it is only used for relativ calls/jumps
      */
	return Addr;
};

/*
 *
 */
void	analyser::nameaddr(ADDR Addr,	char	*n)
{
};

taddr *analysernewaddr(taddr **addrs, ADDR Addr)
{
	if (*addrs) {
		if (Addr < (*addrs)->addr) return analysernewaddr(&(*addrs)->left, Addr);
		if (Addr > (*addrs)->addr) return analysernewaddr(&(*addrs)->right, Addr);
	} else {
		*addrs = (taddr *) calloc(1, sizeof(taddr));
		(*addrs)->addr	= Addr;
	}
	return (*addrs);
};

/*
 *
 */
taddr *analyser::newaddr(ADDR	Addr)
{
	return analysernewaddr(&addrs, Addr);
};

/*
 *
 */
void	analyser::notifyprogress(ADDR	Addr)
{
	// stub
};

/*
 *
 */
void	analyser::pushaddr(ADDR Addr)
{
	if (validcodeaddr(Addr))	{
		dprintf("addr %08lx pushed\n", Addr);
		addr_queue->enqueue((void *) Addr);
     }
};

void countaddrs(taddr *addr, int *c)
{
	if (addr) {
     	countaddrs(addr->right, c);
          (*c)++;
     	countaddrs(addr->left, c);
     }
}

void saveaddrs(htfile *f, taddr *addr)
{
	if (addr) {
     	saveaddrs(f, addr->left);

          f->write(&addr->addr, sizeof(ADDR));
          fstrput(f, addr->label);
          // xrefs
          txref	*xref = addr->xreflist;
          int		count = 0;
          while (xref) {
          	count++;
               xref = xref->next;
          }
          f->write(&count, sizeof(int));
          xref = addr->xreflist;
          while (xref) {
          	f->write(&(xref->addr), sizeof(ADDR));
               f->write(&(xref->xreftype), sizeof(txreftype));
          	xref = xref->next;
          }
          // comments
          tcomment	*com = addr->comments;
          count = 0;
          while (com) {
               count++;
          	com = com->next;
          }
          f->write(&count, sizeof(int));
          com = addr->comments;
          while (com) {
               fstrput(f, com->text);
               com = com->next;
          }
     	f->write(&addr->type, sizeof(taddrtype));

     	saveaddrs(f, addr->right);
     }
}

/*
 *
 */
void analyser::store(htfile *f)
{
     addr_queue->store(f);
     f->write(&ops_parsed, sizeof(int));
     f->write(&active, sizeof(bool));
     f->write(&next_explored, sizeof(ADDR));
     f->write(&first_explored, sizeof(ADDR));
     f->write(&last_explored, sizeof(ADDR));
     explored->store(f);
     int c=0;
     countaddrs(addrs, &c);
     f->write(&c, sizeof(int));
	saveaddrs(f, addrs);
     unasm->store(f);
};

/*
 *
 */
bool	analyser::savetychangeaddr(ADDR Addr)
{
	addr	= Addr;
     return ((addr==next_explored) || (!validcodeaddr(addr))) ? gotoaddr(addr) : true;
};

/*
 *
 */
taddr *analysersearchaddr(taddr *addrs,	ADDR	Addr)
{
	if (addrs) {
		if (Addr < addrs->addr) return analysersearchaddr(addrs->left, Addr);
		if (Addr > addrs->addr) return analysersearchaddr(addrs->right, Addr);
		return addrs;
	} else return NULL;
};

/*
 *
 */
taddr *analyser::searchaddr(ADDR Addr)
{
	return analysersearchaddr(addrs, Addr);
};

/*
 *
 */
void analyser::setactive(bool mode)
{
	if (mode) {
     	if (!active) {
          	active = true;
               some_analyser_active++;
          }
     } else {
     	if (active) {
          	active = false;
               some_analyser_active--;
          }
     }
};


/*
 *
 */
bool	analyser::validaddr(ADDR	Addr, tsectype	action)
{
	// abstract
     return 0;
};

/*
 *
 */
bool	analyser::validcodeaddr(ADDR Addr)
{
	return validaddr(Addr, sccode);
};

/*
 *
 */
bool	analyser::validreadaddr(ADDR Addr)
{
	return validaddr(Addr, scread);
};

/*
 *
 */
bool	analyser::validwriteaddr(ADDR	Addr)
{
	return validaddr(Addr, scwrite);
};

/*
 *
 *		interface only (there's no internal use)
 *
 */

char bufbuf[80];

bool	analyser::getformattedline(ADDR Addr, int line, char *buf, int *length)
{
	/*
      *	WARNING: this function doesn't include any validation of Addr
      *		    so this should be checked by the main program.
      *            If Addr is out of range bufptr() could occasionally fail.
      */
	taddr *a = searchaddr(Addr);
	int level	= 0;

	*length =	0;

     if (line==1) {
     	sprintf(buf, HEX8FORMAT8, Addr);
     } else {
     	sprintf(buf, "   ...  ");
     }

	if (explored->contains(Addr))	{
//		strcpy(buf, "! ");
		sprintf(buf, "%s"HT_EDIT_BYTE_FORMAT" ! ", buf, Addr, '!', tag_coloridx_unselected);
	} else {
//		strcpy(buf, "  ");
		sprintf(buf, "%s"HT_EDIT_BYTE_FORMAT"   ", buf, Addr, ' ', tag_coloridx_unselected);
	}

	if (a) {
          // comments
		tcomment *c = a->comments;
		while (c)	{
			level++;
			if (level==line) {
				strcat(buf, c->text);
				return true;
			};
			c = c->next;
		};
          // label
		if (a->label) {
			level ++;
			if (level==line) {
				strcat(buf, a->label);
				strcat(buf, ":");
				sprintf(buf, "%-30s", buf);
				// xrefs
#define MAX_XREFS 3
				txref *x = a->xreflist;
				if (x) strcat(buf, ";XREF");
				int i=0;
				while (x)	{
					char	xreftypes[] = "rwojc";
					char	buf2[100];
					if (i++<MAX_XREFS) {
		    				sprintf(buf2, " %c"HEX8FORMAT, xreftypes[x->xreftype], x->addr);
						strcat(buf, buf2);
					} else return true;
					x = x->next;
				};
				// xrefsend
				return true;
			};
		};
		// !! evil, worse, devil, stupid, fucking code: !!
				txref *x = a->xreflist;
				int i=0,j=0;
				bool	onetimeshit= false;
				while (x)	{
					char	xreftypes[] = "rwojc";
					char	buf2[100];
					if (level==line) {

						if (!onetimeshit) strcat(buf,	"                            ;XREF");
						onetimeshit = true;
						sprintf(buf2, " %c"HEX8FORMAT, xreftypes[x->xreftype], x->addr);
						strcat(buf, buf2);
						if ((j++)==MAX_XREFS) return true;
					}
					if ((((++i) % MAX_XREFS)==0) && (x->next))	level ++;

					x = x->next;
				};
				if (onetimeshit) return true;
		    		// xrefsend
		// end
	};

	level ++;
	if (level==line) {
		if (validcodeaddr(Addr))	{
			strcat(buf, "  ");
			strcat(buf, unasm->getstring((byte *)bufptr(Addr), 15, mapaddr(Addr), length));
		} else {
          	if (validaddr(Addr, scvalid)) {
				*length =	1;
				sprintf(buf, "%s  db     0x%02x", buf, *(byte *)bufptr(Addr));
               } else {
				*length =	1;
               	strcat(buf, "  db     ??");
               };
			return true;
		};
		return true;
	};

/*     if (a) {
     	txref *x = a->xreflist;
		while (x) {
			level++;
               if (level==line) {
               	char xreftypes[] = "rwojc";
                    char buf2[100];
               	sprintf(buf2, "XREF %c"HEXFORMAT, xreftypes[x->xreftype], x->addr);
                    strcat(buf, buf2);
                    return true;
               };
               x = x->next;
          };
     }*/
	return false;
};

/*
 *                    :KLJDS:
 */

void	unassembler::init(analyser *A)
{
	analy = A;
}

void	unassembler::load(htfile *f)
{
}

void	unassembler::done()
{
}


void	unassembler::store(htfile *f)
{
}


char *unassembler::addr_sym_func(ADDR Addr)
{
     taddr *a = analy->searchaddr(Addr);
     if ((a) && (a->label)){
/*     	switch (a->type) {
          	case dtunknown {
               	return a->label;
               }
          }*/
		return a->label;
     } else {
     	sprintf(bufbuf, HEX8FORMAT/*" "HT_EDIT_CHAR_FORMAT*/, Addr/*, Addr, '#', 8*/);
          return bufbuf;
     };
}


