/*
 *	analy.h
 */

/*	Copyright (C) 1999 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef analy_h
#define analy_h

#include "htunasm.h"
#include <stdio.h>
#include "global.h"
#include "common.h"
#include "codeanaly.h"
#include "dataanaly.h"
#include "stddata.h"
#include "area.h"

extern int some_analyser_active;

class analyser;

class unassembler: public ht_unasm {
public:
	analyser		*analy;
						unassembler():ht_unasm(32) {};
	virtual				~unassembler() {};
	void					init(analyser *A);
     void					load(htfile *f);
	void	virtual			done();

     void	virtual			store(htfile *f);
     char virtual			*addr_sym_func(ADDR Addr);
};

/***************************************************************************/

typedef enum {dtunknown,	dtbyte, dtword, dtdword,	dtstring,	dtunistring, dtarray} taddrtypetype;

struct taddrtype {
	taddrtypetype	type;
	union {
		dword  length;
		// ...?
	};
};

typedef enum {xrefread, xrefwrite,	xrefoffset, xrefjump, xrefcall} txreftype;

struct txref {
	txref	*next;
	ADDR		addr;
	txreftype	xreftype;
};

struct tcomment {
	tcomment	*next;
	char		*text;
};

struct taddr {
	ADDR		addr;
	taddr	*left, *right;
	char		*label;
	txref	*xreflist;
	tcomment	*comments;
     taddrtype	type;
};

// geht besser..
typedef enum {scvalid, scread, scwrite,	screadwrite, sccode, scinitialized} tsectype;

typedef enum {
			brnobranch,						// straight exec. flow
			brjump,
			brreturn,
			brcall,
			brjXX
			} tbranchtype;

typedef enum {	acread, acwrite, acoffset } taccesstype;

struct taccess	{
		bool			indexed;
		bool			write;
		taccesstype 	type;
};

#define IS_FUNC	1
#define IS_LOC		2
struct tlabel {
	int		type;
     taddr     addr;
};

/*
 *	abstract meens function HAS to be overwritten for using this object
 *	stub meens has not neccessary to be overwritten
 *
 */


class codeanalyser;
class dataanalyser;

class analyser: public object	{
public:
	ADDR			addr;
	tqueue		*addr_queue;
	int			ops_parsed;							// for continueing
	bool			active;
	ADDR			next_explored,	first_explored, last_explored;
	tarea		*explored;
	taddr		*addrs;
     codeanalyser	*code;
     dataanalyser	*data;
     unassembler	*unasm;
     tlabel		*labels;
						analyser() {};
	virtual				~analyser() {};
	void					init();
     void					load(htfile *f);
	void	virtual			done();

	bool					addaddrlabel(ADDR Addr, char *prefix);
	void		    			addcomment(ADDR Addr, int line, char *c);
	bool					addlabel(ADDR Addr,	char	*label);
	bool		  			addlabel(ADDR Addr,	char	*label, char **result);
	void		  			addxref(ADDR from, ADDR to, txreftype action);
	void	virtual			beginanalysis(ADDR Addr);
	ADDR	virtual			branchaddr(char *opcode);				// abstract
	void	virtual			*bufptr(ADDR Addr);						// abstract
	bool			  		continueanalysis();
	void	       			continueanalysisat(ADDR Addr);
	void					dataaccess(ADDR Addr, taccess	access);
	void					dobranch(tbranchtype branch, char *opcode, int len);
     void					engage_codeanalyser();
     taddrtypetype virtual	examinedata(ADDR addr);
	void	virtual			examineopcode(char *opcode);				// abstract
	char	virtual			*getaddrlabel(ADDR Addr);
	bool		  			gotoaddr(ADDR Addr);
     void virtual			init_codeanalyser();
     void virtual			init_dataanalyser();
     void virtual			init_unasm();                                // abstract
	tbranchtype virtual		isbranch(char *opcode);					// abstract
	void	virtual			log(char *s);							// stub
	ADDR	virtual			mapaddr(ADDR Addr);						// stub
	void		  			nameaddr(ADDR Addr,	char	*n);
	taddr				*newaddr(ADDR Addr);
	void	virtual			notifyprogress(ADDR	Addr);				// stub
	void					pushaddr(ADDR Addr);
	bool		  			savetychangeaddr(ADDR Addr);
	taddr				*searchaddr(ADDR Addr);
     void                     setactive(bool mode);
     void	virtual			store(htfile *f);
	bool	virtual			validaddr(ADDR	Addr, tsectype	action);		// abstract
	bool					validcodeaddr(ADDR Addr);
	bool		   			validreadaddr(ADDR Addr);
	bool					validwriteaddr(ADDR	Addr);

///  interface only (there's no internal use)

	bool	virtual			getformattedline(ADDR Addr, int line, char *buf, int *length);
};

#define INVALID_ADDR 0xffffffff

#define MAX_OPS_PER_CONTINUE 20

#endif
