/*
 *	pe_analy.cpp
 */

/*	Copyright (C) 1999 Sebastian Biallas (sb@web-productions.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include "global.h"
#include "analy.h"
#include "pe_analy.h"
#include "codeanaly.h"

#include "htapp.h"
#include "htdebug.h"
#include "htpe.h"
#include "htstring.h"

char		buffer[1024];

void	pe_analyser::init(ht_pe_shared_data *Pe_shared, htfile *File)
{
     pe_shared = Pe_shared;
     file = File;

	ADDR entry=pe_shared->pe32header.entrypoint_address+pe_shared->pe32header.image_base;

     analyser::init();

     /////////////

     addcomment(entry, 0, " ");
     addcomment(entry, 0, ";****************************");
     addcomment(entry, 0, ";  program entry point");
     addcomment(entry, 0, ";****************************");

     int export_count=pe_shared->exports.funcs.count();
	for (int i=0; i<export_count; i++) {
		ht_pe_export_function *f=(ht_pe_export_function *)pe_shared->exports.funcs.get(i);
          if (f->byname) {
          	sprintf(buffer, "; exported function %s, ordinal %04x", f->name, f->ordinal);
          } else {
          	sprintf(buffer, "; unnamed exported function, ordinal %04x ", f->ordinal);
          }
     	addcomment(f->function_address, 0, ";********************************************************");
     	addcomment(f->function_address, 0, htstrdup(buffer));
     	addcomment(f->function_address, 0, ";********************************************************");
          pushaddr(f->function_address);
     }

     int import_count=pe_shared->imports.funcs.count();
     for (int i=0; i<import_count; i++) {
		ht_pe_import_function *f=(ht_pe_import_function *)pe_shared->imports.funcs.get(i);
		ht_pe_import_library *d=(ht_pe_import_library *)pe_shared->imports.libs.get(f->dll);
          if (f->byname) {
	          sprintf(buffer, "%s:%s", d->name, f->name.name);
		} else {
	          sprintf(buffer, "%s:%04x", d->name, f->ordinal);
          }
          addlabel(f->function_address, htstrdup(buffer));
     };

     int dimport_count=pe_shared->dimports.funcs.count();
     for (int i=0; i<dimport_count; i++) {
		ht_pe_dimport_function *f=(ht_pe_dimport_function *)pe_shared->dimports.funcs.get(i);
		ht_pe_dimport_library *d=(ht_pe_dimport_library *)pe_shared->dimports.libs.get(f->dll);
          if (f->byname) {
	          sprintf(buffer, "%s:%s", d->name, f->name.name);
		} else {
	          sprintf(buffer, "%s:%04x", d->name, f->ordinal);
          }
          addlabel(f->function_address, htstrdup(buffer));
     };
};


void pe_analyser::load(htfile *f)
{
	analyser::load(f);
};

void	pe_analyser::done()
{
	analyser::done();
};


ADDR pe_analyser::branchaddr(char *opcode)
{
     ADDR		Addr = INVALID_ADDR;
	sscanf(opcode, "%*s %lx", &Addr);
     return Addr;
};

void *pe_analyser::bufptr(ADDR Addr)
{
	dword ofs;
     Addr-=pe_shared->pe32header.image_base;
	if (!pe_rva_to_ofs(&pe_shared->sections, Addr, &ofs)) assert(0);
     file->seek(ofs);
     file->read(buffer, 15);
     return (void*)buffer;
};

byte digit(char c)
{
	if ((c>='0') && (c<='9')) return (c-'0');
     if ((c>='a') && (c<='f')) return (c-'a'+10);
     if ((c>='A') && (c<='F')) return (c-'A'+10);
     return 42;
};

void	pe_analyser::examineopcode(char *opcode)
{
	/* _bad_ code, doesn't even work */
     ADDR	offset = 0;
     if (!opcode) return;
     while (*opcode) {
     	if (digit(*opcode) <= 9) {
          	do {
               	offset *= 16;
                    offset += digit(*opcode++);
               } while ((*opcode) && (digit(*opcode) <= 15));
               taccess a;
               a.indexed = false;
               a.type = acread;
               dataaccess(offset, a);
		     addxref(offset, addr, xrefoffset);
               offset = 0;
               if (!*opcode) return;
          }
          opcode++;
     }
};

void pe_analyser::init_codeanalyser()
{
	analyser::init_codeanalyser();
     code->loaddefs("sign.def");
}

void pe_analyser::init_unasm()
{
     unasm = new unassembler();
     unasm->init(this);
}

tbranchtype pe_analyser::isbranch(char *opcode)
{
	if (*opcode=='j') {
     	if (*(++opcode)=='m') return brjump; else return brjXX;
	} else if (strncmp("call", opcode, 4)==0) {
     	return brcall;
     } else if (strncmp("ret", opcode, 3)==0) {
     	return brreturn;
	} else return brnobranch;
};

void pe_analyser::log(char *msg)
{
//	LOG(msg);
}

bool pe_analyser::validaddr(ADDR Addr, tsectype action)
{
	pe_section_headers *sections=&pe_shared->sections;
	int i;
     Addr-=pe_shared->pe32header.image_base;
     pe_rva_to_section(sections, Addr, &i);
	PE_SECTION_HEADER *s=sections->sections+i;
     if (!pe_rva_is_valid(sections, Addr)) return false;
	switch (action) {
		case scvalid:
               return true;
          case scread:
               return s->characteristics & PE_SCN_MEM_READ;
          case scwrite:
               return s->characteristics & PE_SCN_MEM_WRITE;
          case screadwrite:
               return s->characteristics & PE_SCN_MEM_WRITE;
          case sccode:
               return s->characteristics & PE_SCN_MEM_EXECUTE;
          case scinitialized:
               return s->characteristics & PE_SCN_CNT_INITIALIZED_DATA;
     }
     return false;
}
