/* 
 *	The HTE Editor
 *	htapp.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "analy.h"
#include "htapp.h"
#include "htdialog.h"
#include "htfile.h"
#include "htkeyb.h"
#include "htstring.h"

#include "viewers.h"

#include <stdarg.h>
#include <string.h>

ht_app *app;
screendrawbuf *screen;

#define app_color_normal				palette[1].value

htpalette app_palette[] =
{
	{"application"},
	{"normal", COLOR(C_WHITE, C_BLACK)}
};

#define menu_color_menuline			palette[1].value

htpalette menu_palette[] =
{
	{"menu"},
	{"menuline", COLOR(C_WHITE, C_BLACK)}
};

#define keyline_color_number			palette[1].value
#define keyline_color_active			palette[2].value
#define keyline_color_inactive		palette[3].value

htpalette keyline_palette[] =
{
	{"keyline"},
	{"number", COLOR(C_WHITE, C_BLACK)},
	{"active", COLOR(C_BLACK, C_CYAN)},
	{"inactive", COLOR(C_WHITE, C_CYAN)}
};

htpalette view_manager_listbox_palette[] =
{
	{"view manager listbox"},
	{"normal", COLOR(C_WHITE, C_BLACK)},
	{"selected & focused", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
	{"selected & unfocused", COLOR(C_LIGHT(C_WHITE), C_BLACK)},
	{"scrollbar", COLOR(C_WHITE, C_BLACK)}
};

/*
 *	CLASS ht_menu
 */

ht_menu::ht_menu(bounds *b)
: ht_view(b, hto_ownbuffer, "menuline")
{
	palette=menu_palette;
}

ht_menu::~ht_menu()
{
}

void ht_menu::draw()
{
	buf->fill(size.x, size.y, size.w, size.h, menu_color_menuline);
}

void ht_menu::handlemsg(htmsg *msg)
{
	ht_view::handlemsg(msg);
}

/*
 *	CLASS ht_keyline
 */

ht_keyline::ht_keyline(bounds *b)
: ht_view(b, 0, "keyline")
{
	palette=keyline_palette;
}

ht_keyline::~ht_keyline()
{
}

void ht_keyline::draw()
{
	buf->fill(size.x, size.y, size.w, size.h, keyline_color_active);
	int x=0;
     for (int i=1; i<=10; i++) {
		htmsg msg;
          msg.msg=htm_funcname;
          msg.data_int=i;
		app->sendmsg(&msg);
          if (msg.msg==htm_funcname_return) {
			char *s=msg.data_str;
			printf(x, 0, keyline_color_number, "%d", i<10 ? i : i-10);
    			for (int j=0; j<size.w/10-1; j++) {
				printf(x+j+1, 0, keyline_color_inactive, " ");
     	     }
          	if (s) {
	          	if (s[0]=='~') {
     	     		printf(x+1, 0, keyline_color_inactive, s+1);
          	     } else {
	          		for (int j=0; j<size.w/10-1; j++) {
     	    				printf(x+j+1, 0, keyline_color_active, " ");
			          }
	          		printf(x+1, 0, keyline_color_active, s);
				}
			}
	         	x+=size.w/10;
          }
     }
}

void ht_keyline::handlemsg(htmsg *msg)
{
	ht_view::handlemsg(msg);
}

/*
 *	CLASS ht_app
 */

ht_app::ht_app(bounds *_b) : ht_group(_b, "main application")
{
	bounds b;
/* init timer */
	h0=new_timer();
/* create logwindow */
	getbounds(&b);
     b.y++;
     b.h-=2;

     ht_window *logwindow=new ht_window(&b, "log window", 1);
     b.x++;
     b.y++;
     b.w-=2;
     b.h-=2;
     logviewer=new ht_logviewer(&b, 128);
     logwindow->insert(logviewer);
     insert(logwindow);
/* create menu */
	getbounds(&b);
     b.h=1;
	menu=new ht_menu(&b);
     insert(menu);
/* create keyline */
	getbounds(&b);
     b.y=b.h-1;
     b.h=1;
	keyline=new ht_keyline(&b);
     insert(keyline);

	palette=app_palette;
}

ht_app::~ht_app()
{
	delete_timer(h0);
}

/*debug*/
#define AVG_TIMINGS 20
long long timings[AVG_TIMINGS];
int cur_timing=0;
/**/

void ht_app::draw()
{
/* show draw timings */
/*     if (cur_timing>=AVG_TIMINGS) cur_timing=0;
	timings[cur_timing++]=get_timer_tick(h0);
	long long avg=0;
     for (int i=0; i<AVG_TIMINGS; i++) avg+=timings[i];
     avg=avg/AVG_TIMINGS;
     buf->printf(57, 1, htc_normal, "avg%d: %llu----------", AVG_TIMINGS, avg);*/
/* display flag if analyser(s) active */
	if (some_analyser_active) buf->printf(50, 1, app_color_normal, "ANALY");
}

void ht_app::delete_window(int index)
{
}

int ht_app::findfreewinnumber()
{
     int i=1;
     int free;
	while (1) {
     	ht_view *w=first;
          free=1;
     	while (w) {
          	if (i==w->getnumber()) {
               	free=0;
                    break;
			}
     		w=w->next;
		}
          if (free) return i;
          i++;
     }
     return -1;
}

ht_window *ht_app::findwindowbynumber(int index)
{
     ht_view *w=first;
     while (w) {
          if (index==w->getnumber()) return (ht_window*)w;
     	w=w->next;
     }
     return 0;
}

void ht_app::focus_window(int index)
{
	ht_window *w=findwindowbynumber(index);
     if (w) focus(w);
}

void ht_app::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_draw) {
//		start_timer(h0);
		ht_group::handlemsg(msg);
//	    	stop_timer(h0);
          draw();
          screen->show();
	} else {
		ht_group::handlemsg(msg);
     }
     if (msg->msg==htm_keypressed) {
	     int i=0;
     	switch (msg->data_int) {
          	case K_Alt_9: i++;
	          case K_Alt_8: i++;
     	     case K_Alt_7: i++;
          	case K_Alt_6: i++;
	          case K_Alt_5: i++;
     	     case K_Alt_4: i++;
          	case K_Alt_3: i++;
	          case K_Alt_2: i++;
     	     case K_Alt_1: i++;
          	     focus_window(i);
                    clearmsg(msg);
	               break;
			case K_Alt_0:
				windowlist_popup();
                    clearmsg(msg);
	          	break;
			case K_Return:
               	view_manager_popup();
                    clearmsg(msg);
	          	break;
			case K_Space: {
               	bounds b, c, d, e;
                    getbounds(&b);
                    b.x=b.w/4;
                    b.y=b.h/4;
                    b.w/=2;
                    b.h/=2;
               	ht_dialog *dialog=new ht_dialog(&b, "ht-vision demo");
                    c.x=b.x+2;
                    c.y=b.y+2;
                    c.w=20;
                    c.h=2;
				ht_inputfield *inputfield=new ht_inputfield(&c, "hallihallo0123456789-0");
                    dialog->insert(inputfield);

                    d.x=b.x+2;
                    d.y=b.y+5;
                    d.w=20;
                    d.h=3;
				ht_listbox *listbox=new ht_listbox(&d);
                    listbox->insertstring("string001");
                    listbox->insertstring("sdfsdfsdfsdfg");
                    listbox->insertstring("aset4356");
                    listbox->insertstring("3--aset4356");
                    listbox->insertstring("=///aset4356");
                    dialog->insert(listbox);

                    e.x=b.x+2;
                    e.y=b.y+10;
                    e.w=8;
                    e.h=1;
                    ht_button *button=new ht_button(&e, "Ok", 0);
                    dialog->insert(button);
     	          execute_dialog(dialog);
                    delete dialog;
	               break;
			}
     	}
	}
}

int ht_app::new_window(char *filename)
{
	bounds b;
	getbounds(&b);
     b.y++;
     b.h-=2;
     htrwfile *file=new htrwfile(filename);
     LOG("loading file \"%s\"...", filename);
     ht_window *window=new ht_window(&b, filename, findfreewinnumber());
     b.x++;
     b.y++;
     b.w-=2;
     b.h-=2;
     ht_xgroup *xgroup=new ht_xgroup(&b, filename);

	viewer_int **i=viewer_ints;
     while (*i) {
     	if (((*i)->init) && ((*i)->detect) && ((*i)->detect(file))) {
          	ht_view *v=(*i)->init(&b, file, window);
               v->sendmsg(htm_complete, 0);
			xgroup->insert(v);
          }
          i++;
     }
	window->insert(xgroup);
     insert(window);
     return 1;
}

int ht_app::logf(const char *format, ...)
{
	char buf[1024];
	va_list arg;
     va_start(arg, format);
	int r=vsprintf(buf, format, arg);
     va_end(arg);
     logviewer->insertline(buf);
     htmsg msg;
     msg.type=htm_draw;
     sendmsg(&msg);
     return r;
}

void ht_app::view_manager_popup()
{
     bounds b, c;
     getbounds(&b);
	b.x=b.w/4;
	b.y=b.h/4;
	b.w/=2;
	b.h/=2;
	ht_dialog *dialog=new ht_dialog(&b, "view manager");

/* create listbox */
     c=b;
	c.x++;
	c.y++;
	c.w-=2;
	c.h-=2;
	ht_listbox_button *listbox=new ht_listbox_button(&c, button_ok);
     listbox->setpalette(view_manager_listbox_palette);
/* insert all browsable views */
     ht_collection *structure=new ht_collection();
	int index=0;
     view_manager_dump(current->getfirstchild(), listbox, structure, 0, &index, getcurrent());
/* and seek to the currently selected */
/* FIXME: doesn't do what should anymore because now the currently selected
   view is always also the last one into the list of browsables... */
	listbox->seek(index);
	dialog->insert(listbox);

	if (execute_dialog(dialog)) {
	     ht_listbox_data data;
     	listbox->getdata(&data);
	     ht_data_ptr *p=(ht_data_ptr*)structure->get(data.index);
          focus((ht_view*)p->data);
	}
	delete dialog;
}

void ht_app::view_manager_dump(ht_view *view, ht_listbox *listbox, ht_collection *structure, int depth, int *currenti, ht_view *currentv)
{
	char str[256];
     char *s=str;
     for (int i=0; i<depth; i++) s+=sprintf(s, "  ");
     while (view) {
     	if (view->options & hto_browsable) {
			sprintf(s, "* %s", view->desc);
               structure->add(new ht_data_ptr(view));
               if (view==currentv)
               	*currenti=structure->count()-1;
          	listbox->insertstring(str);
          }
          view_manager_dump(view->getfirstchild(), listbox, structure, depth+1, currenti, currentv);
     	view=view->next;
     }
}

void ht_app::windowlist_popup()
{
}

int htapp_init()
{
	bounds b;
     screen=new screendrawbuf();
     b.x=0;
     b.y=0;
     b.w=screen->size.w;
     b.h=screen->size.h;
	app=new ht_app(&b);
     return 1;
}

void htapp_done()
{
	delete app;
}

