#ifndef __HTDATA_H__
#define __HTDATA_H__

class ht_tree;

#include "global.h"
#include "htdebug.h"

/*
 *	CLASS ht_data
 */

class ht_data {
public:
	ht_data();
     virtual ~ht_data();
};

/*
 *	CLASS ht_data_dword
 */

class ht_data_dword: public ht_data {
public:
     dword data;
	ht_data_dword(dword d);
     virtual ~ht_data_dword();
};

/*
 *	CLASS ht_data_ptr
 */

class ht_data_ptr: public ht_data {
public:
     void *data;
	ht_data_ptr(void *data);
     virtual ~ht_data_ptr();
};

/*
 *	CLASS ht_data_page
 */

#define HT_PAGE_SIZE 64

class ht_data_page: public ht_data {
protected:
	unsigned char buf[HT_PAGE_SIZE];
     unsigned char obuf[HT_PAGE_SIZE];
public:
	int start, end;
	ht_data_page();
	virtual ~ht_data_page();
	void set(int ofs, unsigned char c);
	void seto(int ofs, unsigned char c);
	unsigned char get(int ofs);
	void getbuf(void *buf, int ofs, int size);
     int dirty(int ofs);
};

/*
 *	CLASS ht_tree
 */

struct ht_node {
     int key;
     ht_data *data;
     ht_node *left, *right;
};

class ht_tree {
private:
	int it_key;
     int count;
     int cached_key;
	ht_data *cached_data;
	ht_node *root;
	void set_r(ht_node **node, int key, ht_data *data);
	ht_node *get_r(ht_node *node, int key);
     ht_node *get_next_r(ht_node *node, int key);
     ht_node *get_smallest(ht_node *node);
     void free_r(ht_node *node);
	ht_node *balance_r(ht_node **table, int l, int r);
     void free_skeleton_r(ht_node *n);
public:
	ht_tree();
	~ht_tree();
/* new */
     ht_node *enum_first();
     ht_node *enum_next();
     void set(int key, ht_data *data);
     ht_data *get(int key);
     void del(int key);
     int getcount();

     void balance();
};

/*
 *   ht_collection
 */

class ht_collection {
private:
     ht_data **table;
     int table_size, table_maxsize;

     void expand();
public:
     ht_collection();
     ~ht_collection();
     void add(ht_data *data);
     ht_data *get(int index);
     int count();
};

/*
 *   ht_string_collection
 */

class ht_string_collection {
private:
     char **str_table;
     int str_table_size, str_table_maxsize;

     void expand();
public:
     ht_string_collection();
     ~ht_string_collection();
     void add(char *str);
     char *get(int index);
     void del(int index);
     int count();
};

/*
 *	simple int hash
 */

struct int_hash {
     int value;
     char *desc;
};

char *matchhash(int value, int_hash *hash_table);

#endif /* ! __HTDATA_H__ */

