/* 
 *	The HTE Editor
 *	htdialog.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>

#include "htapp.h"
#include "htdialog.h"
#include "htkeyb.h"
#include "htstring.h"

#define inputfield_color_focused		palette[1].value
#define inputfield_color_unfocused		palette[2].value

htpalette inputfield_palette[] =
{
	{"inputfield"},
	{"focused", COLOR(C_LIGHT(C_WHITE), C_BLUE)},
	{"unfocused", COLOR(C_LIGHT(C_WHITE), C_BLUE)}
};

#define button_color_focused			palette[1].value
#define button_color_unfocused		palette[2].value

htpalette button_palette[] =
{
	{"button"},
	{"focused", COLOR(C_LIGHT(C_WHITE), C_GREEN)},
	{"unfocused", COLOR(C_BLACK, C_GREEN)}
};

#define listbox_color_normal			palette[1].value
#define listbox_color_focused			palette[2].value
#define listbox_color_unfocused		palette[3].value
#define listbox_color_scrollbar		palette[4].value

htpalette listbox_palette[] =
{
	{"listbox"},
	{"normal", COLOR(C_BLACK, C_CYAN)},
	{"selected & focused", COLOR(C_LIGHT(C_WHITE), C_GREEN)},
	{"selected & unfocused", COLOR(C_LIGHT(C_WHITE), C_CYAN)},
	{"scrollbar", COLOR(C_CYAN, C_BLACK)}
};

#define dialog_color_normal			palette[1].value

htpalette dialog_palette[] =
{
	{"dialog"},
	{"normal", COLOR(C_WHITE, C_BLACK)}
};

int execute_dialog(ht_dialog *dialog)
{
	int oldx, oldy;
     screen->getcursor(&oldx, &oldy);
	dialog->setstate(ds_normal);
     dialog->sendmsg(htm_draw, 0);
     screen->show();
	while (dialog->getstate()==ds_normal) {
		if (htkeypressed()) {
			do {
     			dialog->sendmsg(htm_keypressed, htgetkey());
			} while (htkeypressed());
			dialog->sendmsg(htm_draw, 0);
     		screen->show();
          }
	}
     int state=dialog->getstate();
     screen->setcursor(oldx, oldy);
     if (state!=ds_term_cancel) {
     	return state;
	} else {
     	return 0;
     }
}

int msgbox(int buttonmask, char *title, char *format, ...)
{
	char buf[1024];
	va_list arg;
     va_start(arg, format);
	vsprintf(buf, format, arg);
     va_end(arg);
	bounds b, c;
	app->getbounds(&b);
	b.x=b.w/4;
	b.y=b.h/3;
	b.w/=2;
	b.h/=3;
	ht_dialog *dialog=new ht_dialog(&b, title);

     c.x=b.x+2;
     c.y=b.y+1;
     c.w=b.w-4;
     c.h=b.h-3;
     ht_statictext *text=new ht_statictext(&c, buf);
     dialog->insert(text);

     int buttons=0;
     if (buttonmask & msgbox_ok) buttons++;
     if (buttonmask & msgbox_cancel) buttons++;
     if (buttonmask & msgbox_yes) buttons++;
     if (buttonmask & msgbox_no) buttons++;

     int pos=0;
     if (buttonmask & msgbox_ok) {
     	c.x=b.x+b.w/2-5*(buttons-2*pos);
          c.y=b.y+b.h-2;
          c.w=8;
          c.h=1;
     	ht_button *button=new ht_button(&c, "Ok", button_ok);
          dialog->insert(button);
          pos++;
     }
     if (buttonmask & msgbox_yes) {
     	c.x=b.x+b.w/2-5*(buttons-2*pos);
          c.y=b.y+b.h-2;
          c.w=8;
          c.h=1;
     	ht_button *button=new ht_button(&c, "Yes", button_yes);
          dialog->insert(button);
          pos++;
     }
     if (buttonmask & msgbox_no) {
     	c.x=b.x+b.w/2-5*(buttons-2*pos);
          c.y=b.y+b.h-2;
          c.w=8;
          c.h=1;
     	ht_button *button=new ht_button(&c, "No", button_no);
          dialog->insert(button);
          pos++;
     }
     if (buttonmask & msgbox_cancel) {
     	c.x=b.x+b.w/2-5*(buttons-2*pos);
          c.y=b.y+b.h-2;
          c.w=8;
          c.h=1;
     	ht_button *button=new ht_button(&c, "Cancel", button_cancel);
          dialog->insert(button);
          pos++;
     }

	int r=execute_dialog(dialog);
	delete dialog;
     return r;
}

int warning(char *format, ...)
{
	char buf[1024];
	va_list arg;
     va_start(arg, format);
	vsprintf(buf, format, arg);
     va_end(arg);
     return msgbox(msgbox_ok, "warning", buf);
}

int error(char *format, ...)
{
	char buf[1024];
	va_list arg;
     va_start(arg, format);
	vsprintf(buf, format, arg);
     va_end(arg);
     return msgbox(msgbox_ok, "error", buf);
}

/*
 *	CLASS ht_dialog
 */

ht_dialog::ht_dialog(bounds *b, char *desc) : ht_window(b, desc, 0)
{

}

ht_dialog::~ht_dialog()
{
}

void ht_dialog::draw()
{
	clear();
     ht_window::draw();
}

int ht_dialog::getstate()
{
	return state;
}

void ht_dialog::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
	     switch (msg->data_int) {
			case K_Escape:
				setstate(ds_term_cancel);
                    clearmsg(msg);
				break;
	    	}
	} else if (msg->msg==htm_button_pressed) {
          switch (msg->data_int) {
			case button_cancel:
				setstate(ds_term_cancel);
                    clearmsg(msg);
				break;
			default:
				setstate(ds_term_ok);
                    clearmsg(msg);
				break;
          }
     }
	ht_window::handlemsg(msg);
}

void ht_dialog::setstate(int _state)
{
	state=_state;
}

/*
 *	CLASS ht_dialog_element
 */

ht_dialog_element::ht_dialog_element(bounds *b, int options, char *desc)
: ht_view(b, options, "some dialog_element")
{
}

ht_dialog_element::~ht_dialog_element()
{
}

ABSTRACT void ht_dialog_element::getdata(void *buf)
{
	abstract();
}

ABSTRACT int ht_dialog_element::getdatasize()
{
	abstract();
     return 0;
}

/*
 *	CLASS ht_inputfield
 */

ht_inputfield::ht_inputfield(bounds *b, char *deftext)
: ht_dialog_element(b, hto_selectable | hto_ownbuffer,"some inputfield")
{
	text=(char*)malloc(b->w*b->h+1);
     *text=0;
     if (deftext) strcpy(text, deftext);
     curchar=text;
     insert=1;
     palette=inputfield_palette;
}

ht_inputfield::~ht_inputfield()
{
	if (text) delete text;
}

void ht_inputfield::draw()
{
	int c=focused ? inputfield_color_focused : inputfield_color_unfocused;
	buf->fill(size.x, size.y, size.w, size.h, c);
     char *t=text, *u=t;
     int i=0;
     do {
     	u=t;
          t+=lprint(0, i, c, size.w, t);
     	if ((focused) && (curchar>=u) && (curchar<t)) {
          	setcursor(curchar-u, i);
          }
          i++;
     } while ((*t) && (i<size.h));
     if ((focused) && (t==curchar)) setcursor(curchar-u, i-1);
}

void ht_inputfield::getdata(void *buf)
{
	((ht_inputfield_data*)buf)->text=text;
}

int ht_inputfield::getdatasize()
{
	return sizeof text;
}

void ht_inputfield::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
     	switch (msg->data_int) {
          	case K_Up:
               	curchar-=size.w;
                    if (curchar<text) curchar=text;
                    dirtyview();
                    clearmsg(msg);
               	break;
          	case K_Down:
               	curchar+=size.w;
                    if (curchar>text+strlen(text)) curchar=text+strlen(text);
                    dirtyview();
                    clearmsg(msg);
               	break;
          	case K_Left:
                    if (curchar>text) curchar--;
                    dirtyview();
                    clearmsg(msg);
               	break;
          	case K_Right:
                    if ((*curchar) && (curchar-text<size.w*size.h-1)) curchar++;
                    dirtyview();
                    clearmsg(msg);
               	break;
			case K_BackSpace:
               	if (curchar>text) {
                    	curchar--;
               		memmove(curchar, curchar+1, strlen(curchar));
				}
                    dirtyview();
                    clearmsg(msg);
               	break;
			case K_Delete:
                    if (*curchar) memmove(curchar, curchar+1, strlen(curchar));
                    dirtyview();
                    clearmsg(msg);
               	break;
			case K_Home:
               	curchar=text;
                    dirtyview();
                    clearmsg(msg);
               	break;
			case K_End:
               	curchar=text+strlen(text);
                    if (curchar-text>=size.w*size.h) curchar--;
                    dirtyview();
                    clearmsg(msg);
               	break;
			case K_Insert:
               	insert=!insert;
                    dirtyview();
                    clearmsg(msg);
               	break;
			default:
               	if ((msg->data_int>=32) && (msg->data_int<=0xff)) {
               		setchar(msg->data_int, insert);
                         dirtyview();
                         clearmsg(msg);
				}
          }
     }
	ht_dialog_element::handlemsg(msg);
}

void ht_inputfield::setchar(char a, int insert)
{
     if (curchar-text<size.h*size.w) {
		if (insert) {
          	int l=strlen(curchar);
               if (l+curchar-text<size.h*size.w) {
				memmove(curchar+1, curchar, l+1);
				if (!*curchar) *(curchar+1)=0;
				*curchar=a;
				if (curchar-text<size.h*size.w-1) curchar++;
			}
		} else {
			if (!*curchar) *(curchar+1)=0;
			*curchar=a;
			if (curchar-text<size.h*size.w-1) curchar++;
		}
     }
}

/*
 *	CLASS ht_inputfield_button
 */

ht_inputfield_button::ht_inputfield_button(bounds *b, char *deftext, int _value)
: ht_inputfield(b, deftext)
{
	value=_value;
     pressed=0;
}

ht_inputfield_button::~ht_inputfield_button()
{
}

void ht_inputfield_button::getdata(void *buf)
{
	((ht_inputfield_button_data*)buf)->text=text;
	((ht_inputfield_button_data*)buf)->pressed=pressed;
}

int ht_inputfield_button::getdatasize()
{
	return sizeof text+sizeof pressed;
}

void ht_inputfield_button::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
     	switch (msg->data_int) {
			case K_Return:
                    group->sendmsg(htm_button_pressed, value);
                    pressed=1;
                    dirtyview();
                    clearmsg(msg);
               	break;
          }
     }
	ht_inputfield::handlemsg(msg);
}

/*
 *	CLASS ht_button
 */

ht_button::ht_button(bounds *b, char *_text, int _value)
: ht_dialog_element(b, hto_selectable | hto_ownbuffer, "some button")
{
	value=_value;
     text=_text;
     pressed=0;
     palette=button_palette;
}

ht_button::~ht_button()
{
}

void ht_button::draw()
{
	int c=focused ? button_color_focused : button_color_unfocused;
	buf->fill(size.x, size.y, size.w, size.h, c);
	print((size.w-strlen(text))/2, size.h/2, c, text);
}

void ht_button::getdata(void *buf)
{
     ((ht_button_data*)buf)->pressed=pressed;
}

int ht_button::getdatasize()
{
	return sizeof value;
}

void ht_button::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
     	switch (msg->data_int) {
			case K_Return:
                    group->sendmsg(htm_button_pressed, value);
                    pressed=1;
                    dirtyview();
                    clearmsg(msg);
               	break;
		}
     }
	ht_dialog_element::handlemsg(msg);
}

/*
 *	CLASS ht_listbox
 */

ht_listbox::ht_listbox(bounds *b)
: ht_dialog_element(b, hto_selectable | hto_ownbuffer, "some listbox")
{
     list=new ht_string_collection();
	pos=0;
     cursor=0;
     palette=listbox_palette;
}

ht_listbox::~ht_listbox()
{
     delete list;
}

void ht_listbox::draw()
{
     int w=size.w;
     int l=list->count();
     int d=l-pos;
     if (d>size.h) d=size.h;
     int scrollbar=0;
     if (l>size.h) {
          w--;
          scrollbar=1;
     }
	buf->fill(size.x, size.y, size.w, size.h, listbox_color_normal);
	for (int i=0; i<d; i++) {
		int c=(i==cursor) ? (focused ? listbox_color_focused : listbox_color_unfocused) : listbox_color_normal;
          if (scrollbar) {
          	if (i==0) {
               	printf(size.w-1, i, listbox_color_scrollbar, "%c", CHAR_UP);
               } else if (i==d-1) {
               	printf(size.w-1, i, listbox_color_scrollbar, "%c", CHAR_DOWN);
               } else {
               	printf(size.w-1, i, listbox_color_scrollbar, "%c", CHAR_X);
			}
          }
		if (i==cursor) {
			buf->fill(size.x+0, size.y+i, w, 1, c);
          }
     	lprint(0, i, c, w, list->get(i+pos));
     }
}

void ht_listbox::getdata(void *buf)
{
     ((ht_listbox_data*)buf)->list=list;
     ((ht_listbox_data*)buf)->index=pos+cursor;
}

int ht_listbox::getdatasize()
{
	return 4+4;
}

void ht_listbox::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
          	case K_Up:
                    if (cursor) {
                    	cursor--;
				} else {
                         if (pos) pos--;
                    }
                    dirtyview();
                    clearmsg(msg);
               	break;
          	case K_Down:
                    if (cursor+1<size.h) {
                    	if (cursor+1<list->count()) cursor++;
				} else {
                         if (pos+size.h<list->count()) pos++;
                    }
                    dirtyview();
                    clearmsg(msg);
               	break;
          }
     }
	ht_dialog_element::handlemsg(msg);
}

void ht_listbox::insertstring(char *str)
{
     list->add(str);
}

int ht_listbox::seek(int index)
{
	if (index<list->count()) {
	     cursor=index;
     	pos=0;
	     if (cursor+1>size.h) {
     		pos=cursor+1-size.h;
     		cursor=size.h-1;
		}
          return 1;
	}
     return 0;
}

/*
 *	CLASS ht_listbox_button
 */

ht_listbox_button::ht_listbox_button(bounds *b, int _value)
: ht_listbox(b)
{
	value=_value;
}

ht_listbox_button::~ht_listbox_button()
{
}

void ht_listbox_button::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
     	switch (msg->data_int) {
          	case K_Return:
                    group->sendmsg(htm_button_pressed, value);
                    dirtyview();
                    clearmsg(msg);
               	break;
          }
     }
	ht_listbox::handlemsg(msg);
}

/*
 *	CLASS ht_statictext
 */

ht_statictext::ht_statictext(bounds *b, char *_text)
: ht_dialog_element(b, hto_ownbuffer, "some statictext")
{
	text=_text;
     palette=dialog_palette;
}

ht_statictext::~ht_statictext()
{
}

void ht_statictext::draw()
{
	buf->fill(size.x, size.y, size.w, size.h, dialog_color_normal);
     char *t=text, *u=t;
     int i=0;
     do {
     	u=t;
          t+=lprint(0, i, dialog_color_normal, size.w, t);
          i++;
     } while ((*t) && (i<size.h));
}

void ht_statictext::getdata(void *buf)
{
     ((ht_statictext_data*)buf)->text=text;
}

int ht_statictext::getdatasize()
{
	return sizeof text;
}

