#ifndef __HTDIALOG_H__
#define __HTDIALOG_H__

#include "htdebug.h"
#include "htobj.h"

/*
 *	CLASS ht_dialog
 */

#define ds_normal			0
#define ds_term_ok			1
#define ds_term_cancel		2

#define button_cancel		0
#define button_ok			1
#define button_yes            button_ok
#define button_no			2

class ht_dialog: public ht_window {
private:
	int state;
public:
     ht_dialog(bounds *b, char *desc);
     ~ht_dialog();
/* overridden */
	virtual 	void draw();
	virtual	void handlemsg(htmsg *msg);
/* new */
     virtual	int getstate();
	virtual	void setstate(int state);
};

/*
 *	CLASS ht_dialog_element
 */

class ht_dialog_element: public ht_view {
public:
     ht_dialog_element(bounds *b, int options, char *desc);
	virtual ~ht_dialog_element();
/* new */
ABSTRACT virtual void getdata(void *buf);
ABSTRACT virtual int getdatasize();
};

/*
 *	CLASS ht_inputfield
 */

struct ht_inputfield_data {
	char *text;
};

class ht_inputfield: public ht_dialog_element {
protected:
	char *text;
     char *curchar;
     int insert;
public:
     ht_inputfield(bounds *b, char *deftext);
	virtual ~ht_inputfield();
/* overwritten */
	virtual 	void draw();
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void handlemsg(htmsg *msg);
/* new */
     		void setchar(char a, int insert);
};

/*
 *	CLASS ht_inputfield_button
 */

struct ht_inputfield_button_data {
	char *text;
     int pressed;
};

class ht_inputfield_button: public ht_inputfield {
private:
     int value;
     int pressed;
public:
     ht_inputfield_button(bounds *b, char *deftext, int value);
	virtual ~ht_inputfield_button();
/* overwritten */
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_button
 */

struct ht_button_data {
	int pressed;
};

class ht_button: public ht_dialog_element {
private:
	int value;
     int pressed;
     char *text;
public:
     ht_button(bounds *b, char *text, int value);
	virtual ~ht_button();
/* overwritten */
	virtual 	void draw();
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_listbox
 */

struct ht_listbox_data {
	ht_string_collection *list;
	int index;
};

class ht_listbox: public ht_dialog_element {
private:
	ht_string_collection *list;
     int pos, cursor;
public:
     ht_listbox(bounds *b);
	virtual ~ht_listbox();
/* overwritten */
	virtual 	void draw();
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
	virtual	void handlemsg(htmsg *msg);
/* new */
     		void insertstring(char *str);
               int seek(int index);
};

/*
 *	CLASS ht_listbox_button
 */

class ht_listbox_button: public ht_listbox {
private:
	int value;
public:
     ht_listbox_button(bounds *b, int value);
	virtual ~ht_listbox_button();
/* overwritten */
	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_statictext
 */

struct ht_statictext_data {
	char *text;
};

class ht_statictext: public ht_dialog_element {
private:
	char *text;
public:
	ht_statictext(bounds *b, char *text);
	virtual ~ht_statictext();
/* overwritten */
	virtual 	void draw();
	virtual	void getdata(void *buf);
	virtual	int getdatasize();
};

/*
 *	misc
 */

#define msgbox_ok		1
#define msgbox_cancel	2
#define msgbox_yes		4
#define msgbox_no		8

int execute_dialog(ht_dialog *dialog);
int msgbox(int buttonmask, char *title, char *format, ...);
int warning(char *format, ...);
int error(char *format, ...);

#endif /* !__HTDIALOG_H__ */

