#ifndef __HTFILE_H__
#define __HTFILE_H__

#include "htdebug.h"

#include <stdio.h>

/* some stdio.h define getc/putc macros */
#undef getc
#undef putc

/*
 *	CLASS htfile
 */

class htfile {
public:
	htfile();
     virtual ~htfile();
/* new */
	ABSTRACT virtual int read(void *buf, int size);
	ABSTRACT virtual int write(void *buf, int size);
     ABSTRACT virtual int seek(int pos);
     ABSTRACT virtual int tell();
     ABSTRACT virtual int size();
     ABSTRACT virtual char *getname();
     STUB virtual void flush();
	STUB virtual void invd();
     STUB virtual int dirty();
     STUB virtual int bytedirty(int ofs);
};

/*
 *	CLASS htrawfile
 */

class htrawfile: public htfile {
protected:
	char *name, *mode;
     FILE *file;
	int isdirty;
public:

	htrawfile();
     htrawfile(const char *name, const char *mode);
	virtual ~htrawfile();
/* overridden */
	virtual int read(void *buf, int size);
	virtual int write(void *buf, int size);
     virtual int seek(int pos);
     virtual int tell();
	virtual int size();
     virtual char *getname();
     virtual int dirty();
};

/*
 *	CLASS htrwfile
 */

#include "htdata.h"

class htrwfile: public htrawfile {
protected:
	ht_tree *vmod;
     int frag_count;
/* new */
	void putc(unsigned char c);
     unsigned char getc();
     void init_buf();
     void done_buf();
public:
	htrwfile();
     htrwfile(const char *name);
     virtual ~htrwfile();
/* overridden */
	virtual int read(void *buf, int size);
	virtual int write(void *buf, int size);
     virtual int seek(int pos);
     virtual int tell();
     virtual void flush();
	virtual void invd();
     virtual int bytedirty(int ofs);
};

/*
 *	CLASS htmemfile
 */

class htmemfile: public htfile {
protected:
	int ofs;
	int pos;
     int bufsize;
	char *buf;
	int isdirty;

     void extendbuf();
public:
	htmemfile();
     htmemfile(int ofs, int size);
     virtual ~htmemfile();
/* overridden */
	virtual int read(void *buf, int size);
	virtual int write(void *buf, int size);
     virtual int seek(int pos);
     virtual int tell();
	virtual int size();
     virtual char *getname();
     virtual int dirty();
};

/*
 *	string file functions
 */

char *getstrz(htfile *file);
void putstrz(htfile *file, char *str);

char *getstrl(htfile *file);
void putstrl(htfile *file, char *str);

#endif /* !__HTFILE_H__ */
