/* 
 *	The HTE Editor
 *	htobj.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "config.h"

#include "htapp.h"
#include "htcurses.h"
#include "htdebug.h"
#include "htkeyb.h"
#include "htlayout.h"
#include "htmeminf.h"
#include "htobj.h"
#include "htstring.h"
#include "httools.h"
#include "htunasm.h"

#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#define window_color_border	palette[1].value
#define window_color_killer   palette[2].value

htpalette window_palette[] =
{
	{"window"},
	{"border", COLOR(C_WHITE, C_BLACK)},
	{"killer", COLOR(C_WHITE, C_BLACK)}
};

#define viewer_color_normal        palette[1].value
/*#define viewer_color_selected      palette[2].value
#define viewer_color_edited        palette[3].value
#define viewer_color_modified		palette[4].value*/

htpalette viewer_palette[] =
{
	{"viewer"},
	{"normal", COLOR(C_WHITE, C_BLACK)}
};

#define tag_color_normal		palette[tag_coloridx_normal].value
#define tag_color_selected	palette[tag_coloridx_selected].value
#define tag_color_unselected	palette[tag_coloridx_unselected].value
#define tag_color_edited		palette[tag_coloridx_edited].value
#define tag_color_modified	palette[tag_coloridx_modified].value

htpalette tag_palette[] =
{
	{"tags"},
	{"normal", COLOR(C_WHITE, C_BLACK)},
	{"selected", COLOR(C_GREEN, C_BLACK)},
	{"unselected", COLOR(C_LIGHT(C_BLACK), C_BLACK)},
	{"edited", COLOR(C_RED, C_BLACK)},
	{"modified", COLOR(C_CYAN, C_BLACK)}
};

/*
 *	CLASS ht_view
 */

ht_view::ht_view(bounds *b, int _options, char *_desc)
: object()
{
	desc=htstrdup(_desc);
	group=0;
	focused=0;
     view_is_dirty=1;
     size.x=0;
     size.y=0;
     size.w=0;
     size.h=0;
     prev=0;
     next=0;
	options=_options;
     if (options & hto_ownbuffer) {
          buf=new drawbuf(&size);
	} else {
     	buf=screen;
     }
     setbounds(b);
}

ht_view::~ht_view()
{
	if (desc) delete desc;
}

void ht_view::clear()
{
    	buf->fill(size.x, size.y, size.w, size.h, C_WHITE);
}

void ht_view::cleanview()
{
     view_is_dirty=0;
}

void ht_view::clearmsg(htmsg *msg)
{
	msg->msg=htm_nothing;
     msg->type=htmt_nothing;
}

void ht_view::dirtyview()
{
     view_is_dirty=1;
}

STUB void ht_view::draw()
{
}

int ht_view::focus(ht_view *view)
{
	if (view==this) {
          if (!focused) receivefocus();
          return 1;
     }
     return 0;
}

void ht_view::getbounds(bounds *b)
{
	*b=size;
}

ht_view *ht_view::getcurrent()
{
     return this;
}

STUB ht_view *ht_view::getfirstchild()
{
     return 0;
}

int ht_view::getnumber()
{
	return -1;
}

void ht_view::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_draw) {
     	redraw();
	}
}

void ht_view::hidecursor()
{
     screen->hidecursor();
}

int ht_view::isviewdirty()
{
	return view_is_dirty;
}

int ht_view::lprint(int x, int y, int c, int l, char *text)
{
	return buf->lprint(size.x+x, size.y+y, c, l, text);
}

void ht_view::move(int x, int y)
{
     size.x+=x;
     size.y+=y;
}

int ht_view::printf(int x, int y, int c, char *format, ...)
{
	char buf[1024];
	va_list arg;
     va_start(arg, format);
	vsprintf(buf, format, arg);
     va_end(arg);
     return print(x, y, c, buf);
}

int ht_view::print(int x, int y, int c, char *text)
{
	return buf->print(size.x+x, size.y+y, c, text);
}

void ht_view::receivefocus()
{
	dirtyview();
	focused=1;
}

void ht_view::redraw()
{
	if (options & hto_ownbuffer) {
		if (isviewdirty()) {
			draw();
			cleanview();
		}
		screen->drawbuffer((drawbuf*)buf, size.x, size.y);
	} else {
		draw();
	}
}

void ht_view::releasefocus()
{
	dirtyview();
     hidecursor();
	focused=0;
}

void ht_view::sendmsg(htmsg *msg)
{
	handlemsg(msg);
}

void ht_view::sendmsg(int msg, int data)
{
	htmsg m;
     switch (msg) {
     	case htm_nothing:
          	return;
     	case htm_draw:
          	m.msg=htm_draw;
			m.type=htmt_broadcast;
               m.data_int=data;
               break;
		default:
          	m.type=htmt_nothing;
               m.msg=msg;
               m.data_int=data;
               break;
     }
     sendmsg(&m);
}

void ht_view::setbounds(bounds *b)
{
	size=*b;
     if (options & hto_ownbuffer) {
		buf->setbounds(&size);
     }
}

void ht_view::setcursor(int x, int y)
{
     screen->setcursor(size.x+x, size.y+y);
}

void ht_view::setpalette(htpalette *_palette)
{
	palette=_palette;
}

/*
 *	CLASS ht_group
 */

ht_group::ht_group(bounds *b, char *desc)
: ht_view(b, hto_browsable, desc)
{
	first=0;
	current=0;
     last=0;
     shared_data=0;
}

ht_group::~ht_group()
{
	ht_view *a, *b;
     a=first;
     while (a) {
     	b=a->next;
          delete a;
          a=b;
     }
}

STUB int ht_group::focus(ht_view *view)
{
	ht_view *v=first;
     while (v) {
		if (v->focus(view)) {
          	if (current!=v) {
	               current->releasefocus();
                    current=v;
                    putontop(v);
			}
               receivefocus();
			return 1;
          }
     	v=v->next;
     }
     return ht_view::focus(view);
}

ht_view *ht_group::getcurrent()
{
     return current->getcurrent();
}

ht_view *ht_group::getfirstchild()
{
	return first;
}

STUB void ht_group::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
          	case K_Tab: {
                	ht_view *c=current;
                    int s=0;
               	do {
                         if (c->next) {
						c=c->next;
					} else {
                         	c=first;
                              s++;
					}
				} while ((s<2) && (!(c->options & hto_selectable)));
                    if (s<2) {
	                    focus(c);
     	               dirtyview();
          	          clearmsg(msg);
				}
               	break;
			}
          }
     }
	if (msg->type==htmt_broadcast) {
		ht_view::handlemsg(msg);
          ht_view *v=first;
          while (v) {
               v->handlemsg(msg);
               v=v->next;
          }
	} else {
     	current->handlemsg(msg);
		ht_view::handlemsg(msg);
     }
}

void ht_group::insert(ht_view *view)
{
	if (last) last->next=view;
	view->prev=last;
     view->next=0;
	last=view;
     if (current) current->releasefocus();
     current=view;
     current->receivefocus();
     if (!first) first=view;
     view->group=this;
}

void ht_group::putontop(ht_view *view)
{
	if (view->next) {
		if (view->prev) view->prev->next=view->next; else first=view->next;
		view->next->prev=view->prev;
          view->prev=last;
          view->next=0;
          last->next=view;
          last=view;
	}
}

STUB void ht_group::redraw()
{
	ht_view::redraw();
     ht_view *v=first;
     while (v) {
          v->redraw();
	     v=v->next;
	}
}

/*
 *	CLASS ht_xgroup
 */

ht_xgroup::ht_xgroup(bounds *b, char *desc) : ht_group(b, desc)
{
	first=0;
	current=0;
     last=0;
}

ht_xgroup::~ht_xgroup()
{
}

STUB void ht_xgroup::handlemsg(htmsg *msg)
{
/* no broadcasts. */
    	if (current) current->handlemsg(msg);
	ht_view::handlemsg(msg);
}

/*
 *	CLASS ht_winborder
 */

ht_winborder::ht_winborder(bounds *b, char *desc, int _number)
: ht_view(b, 0, desc)
{
     number=_number;
     palette=window_palette;
}

ht_winborder::~ht_winborder()
{
}

STUB void ht_winborder::draw()
{
	char buf[1024];
     char *s;
	bounds b;
     getbounds(&b);
/* "...Ŀ" */
	s=buf;
	s+=sprintf(s, "%c", CHAR_CORNERUL);
     for (int i=0; i<b.w-2; i++) s+=sprintf(s, "%c", CHAR_BORDERH);
	s+=sprintf(s, "%c", CHAR_CORNERUR);
     print(0, 0, window_color_border, buf);
/* "..." */
	s=buf;
	s+=sprintf(s, "%c", CHAR_CORNERLL);
     for (int i=0; i<b.w-2; i++) s+=sprintf(s, "%c", CHAR_BORDERH);
	s+=sprintf(s, "%c", CHAR_CORNERLR);
     print(0, b.h-1, window_color_border, buf);
/* "", "" */
	sprintf(buf, "%c", CHAR_BORDERV);
	for (int i=1; i<b.h-1; i++) {
		print(0, i, window_color_border, buf);
		print(b.w-1, i, window_color_border, buf);
	}
/* "[x]" */
	print(2, 0, window_color_border, "[ ]");
	print(3, 0, window_color_killer, "x");
/* <title> */
	if (desc) {
	     int l=strlen(desc), o=0;
     	if (l>b.w-15) {
	     	o=l-b.w+15;
     	     l-=o;
	    		printf((b.w-l)/2, 0, window_color_border, " ...%s ", desc+o);
		} else {
     		printf((b.w-l)/2, 0, window_color_border, " %s ", desc);
		}
     }
/* e.g. "1" */
	if (number) {
		int l=number;
     	int i=0;
	     do {
     		l=l/10;
          	i++;
	     } while (l);
		printf(b.w-2-i, 0, window_color_border, "%d", number);
	}
}

STUB void ht_winborder::handlemsg(htmsg *msg)
{
     if (msg->msg==htm_keypressed) {
/*		switch (msg->data_int) {
     		case K_Control_F5:
               	dirtyview();
				resize=!resize;
                    clearmsg(msg);
          		break;
		}*/
	}
     ht_view::handlemsg(msg);
}

/*
 *	CLASS ht_window
 */

ht_window::ht_window(bounds *b, char *desc, int _number) : ht_group(b, desc)
{
	border=new ht_winborder(b, desc, _number);
     insert(border);
     palette=window_palette;
     number=_number;
}

ht_window::~ht_window()
{
}

int ht_window::getnumber()
{
	return number;
}

/*
 *	CLASS ht_viewer
 */

ht_viewer::ht_viewer(bounds *b, char *desc, htfile *_file)
:ht_view(b, hto_ownbuffer | hto_browsable | hto_selectable, desc)
{
     file=_file;
	shared_data=0;
     palette=viewer_palette;
}

ht_viewer::~ht_viewer()
{
}

STUB char *ht_viewer::func(int i, int getname)
{
     return 0;
}

STUB int ht_viewer::jump(ADDR ofs)
{
	return 0;
}

STUB int ht_viewer::jumpstr(char *str)
{
	char *e;
	int v=strtol(str, &e, 16);
     if (e==str) return 0;
	return jump(v);
}

STUB void ht_viewer::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
     	int i=0;
     	switch (msg->data_int) {
			case K_F12: i++;
			case K_F11: i++;
			case K_F10: i++;
			case K_F9: i++;
			case K_F8: i++;
			case K_F7: i++;
			case K_F6: i++;
			case K_F5: i++;
			case K_F4: i++;
			case K_F3: i++;
			case K_F2: i++;
			case K_F1: i++;
               	sendmsg(htm_funcexec, i);
                    clearmsg(msg);
	               break;
			case K_Up:
               	cursor_up(1);
                    clearmsg(msg);
                    dirtyview();
	               break;
			case K_Down:
          	     cursor_down(1);
                    clearmsg(msg);
                    dirtyview();
	               break;
			case K_Left:
     	          cursor_left(1);
                    clearmsg(msg);
                    dirtyview();
	               break;
			case K_Right:
     	          cursor_right(1);
                    clearmsg(msg);
                    dirtyview();
	               break;
			case K_PageUp: {
     	     	bounds b;
          	     getbounds(&b);
               	cursor_up(b.h-1);
                    clearmsg(msg);
                    dirtyview();
                    break;
			}
			case K_PageDown: {
          		bounds b;
	               getbounds(&b);
     	          cursor_down(b.h-1);
                    clearmsg(msg);
                    dirtyview();
                    break;
			}
			case K_Control_Left: {
          		scroll_left(1);
                    clearmsg(msg);
                    dirtyview();
                    break;
			}
			case K_Control_Right: {
	          	scroll_right(1);
                    clearmsg(msg);
                    dirtyview();
                    break;
			}
          }
	} else if (msg->msg==htm_funcexec) {
		func(msg->data_int, 0);
	} else if (msg->msg==htm_funcname) {
          msg->msg=htm_funcname_return;
		msg->data_str=func(msg->data_int, 1);
     } else if (msg->msg==htm_complete) {
     	complete();
     }
	ht_view::handlemsg(msg);
}

STUB void ht_viewer::complete()
{
}

STUB void ht_viewer::scroll_up(int n)
{
}

STUB void ht_viewer::scroll_down(int n)
{
}

STUB void ht_viewer::scroll_left(int n)
{
}

STUB void ht_viewer::scroll_right(int n)
{
}

STUB int ht_viewer::cursor_up(int n)
{
	return 0;
}

STUB int ht_viewer::cursor_down(int n)
{
	return 0;
}

STUB int ht_viewer::cursor_left(int n)
{
	return 0;
}

STUB int ht_viewer::cursor_right(int n)
{
	return 0;
}

/*
 *	CLASS ht_uviewer
 */

ht_uviewer::ht_uviewer(bounds *b, char *_desc, htfile *_file)
: ht_viewer(b, _desc, _file)
{
	firstsub=0;
	currentsub=0;
	lastsub=0;
     ofs=0;
     yofs=0;
     xofs=0;
     cursor_sub=0;
     cursor_eofs=0;
     cursor_ofs=0;
     cursor_yofs=0;
     cursor_group=0;
     cursor_e=0;
     cursor_x=0;
     cursor_y=0;
     edit=0;
}

ht_uviewer::~ht_uviewer()
{
	hts_sub *s=firstsub, *t;
     while (s) {
     	t=s->next;
          delete s;
		s=t;
     }
}

char *ht_uviewer::func(int i, int getname)
{
     switch (i) {
     	case 3:
          	if (edit) {
               	if (getname) return "view"; else stop_edit();
			} else {
               	if (getname) return "edit"; else start_edit();
			}
               break;
		case 5:
              	if (getname) return "goto"; else {
                    popup_jump_dialog();
			}
               break;
		case 9:
          	if (edit) {
          		if (getname) return "update"; else update_edit();
			} else {
          		if (getname) return "~update";
			}
               break;
     }
     return 0;
}

void ht_uviewer::handlemsg(htmsg *msg)
{
	if (msg->msg==htm_keypressed) {
		switch (msg->data_int) {
     		case K_Home:
          		cursor_home();
                    clearmsg(msg);
                    dirtyview();
                    break;
	     	case K_End:
     	     	cursor_end();
                    clearmsg(msg);
                    dirtyview();
                    break;
			case K_Tab:
     	     	cursor_next_group();
                    clearmsg(msg);
                    dirtyview();
                    break;
			case K_Escape:
     	          if (edit) {
          	          stop_edit();
                         clearmsg(msg);
                    	dirtyview();
     	          }
                    break;
			default:
               	if (edit) {
                         if (editinput(msg->data_int)) clearmsg(msg);
                    }
        			break;
		}
     }
     ht_viewer::handlemsg(msg);
}

int ht_uviewer::editinput(int key)
{
	char *e=get_editable(curline, cursor_e, cursor_group);
    	switch (e[1]) {
		case HT_EDIT_BYTE: {
			char *p;
			char *hexchars="0123456789abcdef";
			if ((key<0x80) && (p=strchr(hexchars, key))) {
                   	int r=p-hexchars;
                    int o=tag_get_offset(e);
                    int v=0;
                    file->seek(o);
                    file->read((unsigned char*)&v, 1);

  				if (cursor_x) {
                        	v=(v&0xf0)|r;
                    } else {
                         v=(v&0x0f)|(r<<4);
                    }
                    file->seek(o);
                    file->write((unsigned char*)&v, 1);
                    qdraw();
				cursor_right(1);
                    dirtyview();
                    return 1;
			}
               break;
		}
          case HT_EDIT_WORD: {
             	char *p;
              	char *hexchars="0123456789abcdef";
               if ((key<0x80) && (p=strchr(hexchars, key))) {
                   	int r=p-hexchars;
                    int o=tag_get_offset(e);
                    int v=0;
                    file->seek(o);
                    file->read((unsigned char*)&v, 2);

                    switch (cursor_x) {
                        	case 0:
                        		v=(v&0x0fff)|(r<<12);
                              break;
                        	case 1:
                            	v=(v&0xf0ff)|(r<<8);
                            	break;
                        	case 2:
                            	v=(v&0xff0f)|(r<<4);
                             	break;
                       	case 3:
                             	v=(v&0xfff0)|r;
                            	break;
				}
                    file->seek(o);
                    file->write((unsigned char*)&v, 2);
                    qdraw();
				cursor_right(1);
                    dirtyview();
                   	return 1;
			}
			break;
		}
          case HT_EDIT_DWORD: {
              	char *p;
               char *hexchars="0123456789abcdef";
               if ((key<0x80) && (p=strchr(hexchars, key))) {
                   	int r=p-hexchars;
                    int o= tag_get_offset(e);
                    int v=0;
                    file->seek(o);
                    file->read((unsigned char*)&v, 4);

                    switch (cursor_x) {
                       	case 0:
                            	v=(v&0x0fffffff)|(r<<28);
                            	break;
                        	case 1:
                            	v=(v&0xf0ffffff)|(r<<24);
                            	break;
                       	case 2:
                            	v=(v&0xff0fffff)|(r<<20);
                             	break;
                        	case 3:
                             	v=(v&0xfff0ffff)|(r<<16);
                            	break;
                      	case 4:
                             	v=(v&0xffff0fff)|(r<<12);
                             	break;
                         case 5:
                              v=(v&0xfffff0ff)|(r<<8);
                              break;
                         case 6:
                              v=(v&0xffffff0f)|(r<<4);
                              break;
                         case 7:
                            	v=(v&0xfffffff0)|r;
                            	break;
				}
                    file->seek(o);
                    file->write((unsigned char*)&v, 4);
                    qdraw();
				cursor_right(1);
                    dirtyview();
                    return 1;
			}
               break;
		}
          case HT_EDIT_CHAR:
             	if ((key>=' ') && (key<0x100)) {
              		int o= tag_get_offset(e);
                    file->seek(o);
                    file->write((unsigned char*)&key, 1);
                    qdraw();
				cursor_right(1);
                    dirtyview();
                    return 1;
			}
               break;
	}
	return 0;
}

void ht_uviewer::insertsub(hts_sub *sub)
{
	if (lastsub) lastsub->next=sub;
     sub->prev=lastsub;
     lastsub=sub;
     if (!firstsub) firstsub=sub;
     if (!currentsub) currentsub=sub;
     if (!lastsub) lastsub=sub;
}

void ht_uviewer::complete()
{
	firstsub->start_line_address(&ofs, &yofs);
	ADDR _ofs=ofs;
     hts_sub *_sub=firstsub;
     int _yofs=yofs;
     int y=0;
	char linebuf[1024];
    	char *str=linebuf;
     cursor_group=0;
     cursor_sub=0;
     cursor_x=0;
     cursor_y=0;
     while (1) {
     	if (!_sub->getline(str, _ofs, _yofs)) assert(0);
          int c=count_editables(str);
          if (c) {
          	cursor_eofs=tag_get_offset(get_editable(str, 0, cursor_group));
               cursor_ofs=_ofs;
               cursor_yofs=_yofs;
               cursor_sub=_sub;
               cursor_y=y;
               break;
          }
          y++;
          if (!seek_next_line(&_sub, &_ofs, &_yofs, 1)) ERROR("no editables in sub-viewers");
	}
}

void ht_uviewer::apply_modifier(char *htstring)
{
	while ((htstring=findtag(htstring))) {
    		switch (htstring[1]) {
              	case HT_EDIT_BYTE: {
                    dword o=tag_get_offset(htstring);
                    if (file->bytedirty(o)) {
	               	tag_set_color(htstring, tag_coloridx_modified);
				}
				htstring+=HT_EDIT_BYTE_LEN;
				break;
			}
              	case HT_EDIT_WORD: {
                    dword o=tag_get_offset(htstring);
                    if ((file->bytedirty(o)) || (file->bytedirty(o+1))) {
	               	 tag_set_color(htstring, tag_coloridx_modified);
				}
				htstring+=HT_EDIT_WORD_LEN;
				break;
			}
              	case HT_EDIT_DWORD: {
                    dword o=tag_get_offset(htstring);
                    if ((file->bytedirty(o)) || (file->bytedirty(o+1)) || (file->bytedirty(o+2)) || (file->bytedirty(o+3))){
	               	 tag_set_color(htstring, tag_coloridx_modified);
				}
				htstring+=HT_EDIT_DWORD_LEN;
				break;
			}
			case HT_EDIT_CHAR: {
                    dword o=tag_get_offset(htstring);
                    if (file->bytedirty(o)) {
	               	 tag_set_color(htstring, tag_coloridx_modified);
				}
				htstring+=HT_EDIT_CHAR_LEN;
				break;
			}
			default:
				htstring+=2;
               	break;
		}
     }
}

int ht_uviewer::jump(ADDR tofs)
{
	char line[1024];
	char *str=line;
	int y;
     ADDR o;
	hts_sub *s=firstsub;
     while (s) {
     	if (s->jump(tofs, &o, &y)) {
          	while (1) {
				s->getline(str, o, y);
     	     	char *e=get_editable(str, 0, cursor_group);
                    if (e) break;
                    if (!seek_next_line(&s, &o, &y, 1)) return 0;
			}
               strcpy(curline, str);
          	currentsub=s;
               cursor_sub=s;
               cursor_e=0;
               cursor_group=0;
               cursor_x=0;
               cursor_y=0;
               ofs=o;
               yofs=y;
               cursor_ofs=o;
               cursor_yofs=y;
               update_cursor_data();
			dirtyview();
     		return 1;
		}
     	s=s->next;
     }
     return 0;
}

void ht_uviewer::popup_jump_dialog()
{
	bounds b, c;
     app->getbounds(&b);
     b.x=b.w/2-12;
     b.y=b.h/2-2;
     b.w=24;
     b.h=3;
     ht_dialog *dialog=new ht_dialog(&b, "jump");
     c.x=b.x+b.w/2-10;
     c.y=b.y+1;
     c.w=20;
     c.h=1;
     ht_inputfield_button *inputfield=new ht_inputfield_button(&c, 0, button_ok);
     dialog->insert(inputfield);
     if (execute_dialog(dialog)) {
		ht_inputfield_button_data data;
          inputfield->getdata((void*)&data);
     	jumpstr(data.text);
     }
     delete dialog;
}

void ht_uviewer::qdraw()
{
	int cursornotfound=0;
     bounds b;
	getbounds(&b);
	char linebuf[1024];
	char *str=linebuf;
restart:
	clear();
	ADDR _ofs=ofs;
	int _yofs=yofs;
	hts_sub *_sub=currentsub;
     int foundcursor=0;
     for (int y=0; y<b.h; y++) {
		if (!_sub->getline(str, _ofs, _yofs)) assert(0);
          if (!foundcursor) {
          	if ((cursornotfound) || (y==cursor_y)) {
	          	int c=count_editables(str);
     	     	char *e=get_editable(str, 0, cursor_group);
         			if (_sub==cursor_sub) {
          			for (int i=0; i<c; i++) {
               			ADDR o=tag_get_offset(e);
		               	if (o==cursor_eofs) {
found:
     			               cursor_ofs=_ofs;
          			          cursor_yofs=_yofs;
	               		     cursor_e=i;
			                    cursor_y=y;
     			          	strcpy(curline, str);
							update_cursor_data();
		     	               foundcursor=1;
                                   return;
	     	     	     }
	     	     		e=get_editable(e, 1, 0);
					}
	     	     }
                    if ((cursor_y==y) && (!foundcursor)) {
                    	if (cursornotfound) {
                         	e=get_editable(str, 1, 0);
                              if (!e) assert(0);
      					goto found;
                         } else {
                         	cursornotfound=1;
                         	goto restart;
					}
                    }
			}
		}

          if (!seek_next_line(&_sub, &_ofs, &_yofs, 1)) break;
     }
	if (!foundcursor) {
          assert(0);	/* FIXME: debug */
/*
     	scroll_down(1);
          goto restart;*/
     }
}

void ht_uviewer::draw()
{
	int cursornotfound=0;
     bounds b;
	getbounds(&b);
	char linebuf[1024];
	char *str=linebuf;
restart:
	clear();
	ADDR _ofs=ofs;
	int _yofs=yofs;
	hts_sub *_sub=currentsub;
     int foundcursor=0;
     for (int y=0; y<b.h; y++) {
		if (!_sub->getline(str, _ofs, _yofs)) assert(0);
          if (edit) apply_modifier(str);
          if (!foundcursor) {
          	if ((cursornotfound) || (y==cursor_y)) {
	          	int c=count_editables(str);
     	     	char *e=get_editable(str, 0, cursor_group);
         			if (_sub==cursor_sub) {
          			for (int i=0; i<c; i++) {
               			ADDR o=tag_get_offset(e);
		               	if (o==cursor_eofs) {
found:
     			               cursor_ofs=_ofs;
          			          cursor_yofs=_yofs;
	               		     cursor_e=i;
			                    cursor_y=y;
     			          	strcpy(curline, str);
							update_cursor_data();
                                   if (tag_is_editable(e)) {
				               	if (edit) {
			          				int len=htstrlen(str);
     				          		setcursor(len-htstrlen(e)+cursor_x, y);
	          	                    	tag_set_color(e, tag_coloridx_edited);
			     	          	} else {
	     	                         	tag_set_color(e, tag_coloridx_selected);
								}
                                   } else {
     	                         	tag_set_color(e, tag_coloridx_selected);
							}
		     	               foundcursor=1;
	     		          	break;
	     	     	     }
	     	     		e=get_editable(e, 1, 0);
					}
	     	     }
                    if ((cursor_y==y) && (!foundcursor)) {
                    	if (cursornotfound) {
                         	e=get_editable(str, 1, 0);
                              if (!e) assert(0);
      					goto found;
                         } else {
                         	cursornotfound=1;
                         	goto restart;
					}
                    }
			}
		}

		htprint(0, y, b.w, xofs*4, str);
//          drawbufprint(buf, b.x, b.y+y, htc_normal, str);
//			rlprint(b.x, b.y+y, htc_normal, b.w, str);
          if (!seek_next_line(&_sub, &_ofs, &_yofs, 1)) break;
     }
	if (!foundcursor) {
          assert(0);	/* FIXME: debug */
/*
     	scroll_down(1);
          goto restart;*/
     }
     printf(0, b.h, viewer_color_normal, "%c, mem=%d, x=%d, y=%d, csub=%d, cofs=%08x, cyofs=%d, ceofs=%08x, ce=%d", edit ? 'E' : 'V', get_free_mem(), cursor_x, cursor_y, cursor_sub, cursor_ofs, cursor_yofs, cursor_eofs, cursor_e);
}

int ht_uviewer::htprint(int x, int y, int maxlen, int ofs, char *text)
{
 	char *b=text, *r=b;
     htpalette *palette=tag_palette;
	int c=tag_color_normal, cx=x, cy=y, n=0, d, lastv=0;
     while ((r=findtag(b)) && (n<maxlen)) {
          if (ofs<0) ofs=0;
     	if (r!=b) {
     		d=lprint(cx, cy, c, MIN(r-b-ofs, maxlen-n), b+ofs);
               n+=d;
          	cx+=d;
               if (r-b>0) ofs-=r-b;
		}
          if (ofs<0) ofs=0;
     	if (r[0]=='\e') {
          	switch ((unsigned char)r[1]) {
                    case HT_EDIT_BYTE: {
                         int c=palette[hexb(r+12)].value;
                         int v=hexb(r+10);
                         lastv=v;
                         d=lprint(cx, cy, c, MIN(2-ofs, maxlen-n), r+10+ofs);
                         n+=d;
                         cx+=d;
                         ofs-=2;
                         r+=HT_EDIT_BYTE_LEN;
                         break;
				}
                    case HT_EDIT_WORD: {
                         int c=palette[hexb(r+14)].value;
                         int v=hexw(r+10);
                         lastv=v;
                        	d=lprint(cx, cy, c, MIN(4-ofs, maxlen-n), r+10+ofs);
                         n+=d;
                         cx+=d;
                         ofs-=4;
                         r+=HT_EDIT_WORD_LEN;
                         break;
				}
                    case HT_EDIT_DWORD: {
                         int c=palette[hexb(r+18)].value;
                         int v=hexd(r+10);
                         lastv=v;
                         d=lprint(cx, cy, c, MIN(8-ofs, maxlen-n), r+10+ofs);
                         n+=d;
                         cx+=d;
                         ofs-=8;
                         r+=HT_EDIT_DWORD_LEN;
                         break;
				}
                    case HT_EDIT_TIME: {
//                         int c=palette[hexb(r+18)].value;
                         int v=hexd(r+10);
                         lastv=v;
/*					tm *time=gmtime((time_t *)&v);
                         char line[32];
                         sprintf(line, "%02d:%02d:%02d %02d.%02d.%04d", time->tm_hour, time->tm_min, time->tm_sec, time->tm_mday, time->tm_mon+1, time->tm_year+1900);
                         d=lprint(cx, cy, c, MIN(19-ofs, maxlen-n), line+ofs);*/
                         d=2;
                         n+=d;
                         cx+=d;
                         ofs-=8;
                         r+=HT_EDIT_TIME_LEN;
                         break;
				}
                    case HT_EDIT_CHAR: {
                         int c=palette[hexb(r+12)].value;
                         int v=hexb(r+10);
                         lastv=v;
                    	char lb[2];
                         lb[0]=v;
/*                         if (!lb[0]) lb[0]=zerofillchar;*/
                         if (lb[0]<' ') lb[0]=zerofillchar;
                         lb[1]=0;
                         d=lprint(cx, cy, c, MIN(1-ofs, maxlen-n), lb+ofs);
                         n+=d;
                         cx+=d;
                         ofs-=1;
                         r+=HT_EDIT_CHAR_LEN;
                         break;
				}
                    case HT_SEL: {
                         int c=palette[hexb(r+12)].value;
                         int l=hexb(r+10);
					d=lprint(cx, cy, c, MIN(l-ofs, maxlen-n), r+HT_SEL_LEN+ofs);
					n+=d;
					cx+=d;
                         ofs-=1;
                         r+=l+HT_SEL_LEN;
                         break;
                    }
				case HT_EXPL: {
                         int i=hexd(r+2);
            			int_hash *h=find_expl(i);
                         char *s;
                         if ((h) && (s=matchhash(lastv, h))) {
                         	int l=strlen(s);
						d=lprint(cx, cy, c, MIN(l-ofs, maxlen-n), s+ofs);
						n+=d;
						cx+=d;
                			ofs-=l;
                         } else {
						d=lprint(cx, cy, c, MIN(3-ofs, maxlen-n), (char*)"???"+ofs);
						n+=d;
						cx+=d;
                			ofs-=3;
                         }
                         r+=HT_EXPL_LEN;
                    	break;
				}
				case HT_QUOTE:
                    	break;
                    case HT_GROUP:
                         r+=HT_GROUP_LEN;
                    	break;
				default:
                    	assert(0);
                    	break;
			}
          }
          b=r;
     }
     if (ofs<0) ofs=0;
     int l=strlen(b);
     if (n<maxlen) n+=lprint(cx, cy, c, MIN(l-ofs, maxlen-n), b+ofs);
     return n;
}

void ht_uviewer::update_cursor_data()
{
     cursor_group=MIN(cursor_group, count_groups(curline)-1);
     cursor_e=MIN(cursor_e, count_editables_in_group(curline, cursor_group)-1);
     char *e=get_editable(curline, cursor_e, cursor_group);
     cursor_eofs=tag_get_offset(e);
     if (edit) {
     	cursor_x=MIN(cursor_x, tag_get_vlen(e)-1);
	} else {
		cursor_x=0;
	}
}

void ht_uviewer::cursor_home()
{
     cursor_e=0;
     cursor_x=0;
     update_cursor_data();
}

void ht_uviewer::cursor_end()
{
	cursor_e=count_editables_in_group(curline, cursor_group)-1;
    	cursor_x= tag_get_vlen(get_editable(curline, cursor_e, cursor_group))-1;
     update_cursor_data();
}

void ht_uviewer::cursor_next_group()
{
     int g=count_groups(curline);
     cursor_group++;
     if (cursor_group>=g) cursor_group=0;
     update_cursor_data();
}

/*int ht_uviewer::funcapi(int func, int cmd)
{
     switch (func) {
     	FUNC(3, edit ? stop_edit() : start_edit(), edit ? "view" : "edit")
     	FUNC(9, edit ? update_edit() : 0, edit ? "update" : "~update")
     }
     return 0;
}*/

int ht_uviewer::start_edit()
{
	edit=1;
     dirtyview();
     return 1;
}

int ht_uviewer::update_edit()
{
     if (file) {
     	file->flush();
          dirtyview();
	}
     return 1;
}

int ht_uviewer::stop_edit()
{
	edit=0;
     if (file) {
     	file->invd();
          dirtyview();
	}
//     if (!edit) hidecursor();
     return 1;
}

int ht_uviewer::seek_prev_line(hts_sub **sub, ADDR *ofs, int *yofs, int n)
{
	int o=n;
	while (n) {
		if (n-=(*sub)->prev_line_address(ofs, yofs, n)) {
			if (!(*sub)->prev) break;
			*sub=(*sub)->prev;
			(*sub)->end_line_address(ofs, yofs);
			n--;
		}
	}
     return o-n;
}

int ht_uviewer::seek_next_line(hts_sub **sub, ADDR *ofs, int *yofs, int n)
{
	int o=n;
	while (n) {
		if (n-=(*sub)->next_line_address(ofs, yofs, n)) {
			if (!(*sub)->next) break;
			*sub=(*sub)->next;
			(*sub)->start_line_address(ofs, yofs);
			n--;
		}
	}
     return o-n;
}

int ht_uviewer::seek_prev_editable_line(hts_sub **sub, ADDR *ofs, int *yofs)
{
	ADDR o=*ofs;
     int y=*yofs, n=0;
     hts_sub *s=*sub;
	do {
		if (!seek_prev_line(&s, &o, &y, 1)) return 0;
		s->getline(curline, o, y);
          n++;
	} while (!count_editables(curline));
	*ofs=o;
	*yofs=y;
	*sub=s;
     return n;
}

int ht_uviewer::seek_next_editable_line(hts_sub **sub, ADDR *ofs, int *yofs)
{
	ADDR o=*ofs;
	int y=*yofs, n=0;
     hts_sub *s=*sub;
	do {
		if (!seek_next_line(&s, &o, &y, 1)) return 0;
		s->getline(curline, o, y);
          n++;
	} while (!count_editables(curline));
	*ofs=o;
	*yofs=y;
	*sub=s;
     return n;
}

int ht_uviewer::count_lines_to_top(hts_sub *sub, ADDR ofs, int yofs)
{
	int c=0;
	while (seek_prev_line(&sub, &ofs, &yofs, 1)) c++;
	return c;
}

int ht_uviewer::count_lines_to_bottom(hts_sub *sub, ADDR ofs, int yofs)
{
	int c=0;
	while (seek_next_line(&sub, &ofs, &yofs, 1)) c++;
	return c;
}

void ht_uviewer::scroll_up(int n)
{
	if (!seek_prev_line(&currentsub, &ofs, &yofs, n)) assert(0);
}

void ht_uviewer::scroll_down(int n)
{
	if (!seek_next_line(&currentsub, &ofs, &yofs, n)) assert(0);
}

void ht_uviewer::scroll_left(int n)
{
	if (xofs>0) xofs--;
}

void ht_uviewer::scroll_right(int n)
{
	xofs++;
}

int ht_uviewer::cursor_up(int n)
{
	int s=0;
	n=seek_prev_line(&cursor_sub, &cursor_ofs, &cursor_yofs, n);
     if (n) {
		int scrolltop=0;
		cursor_sub->getline(curline, cursor_ofs, cursor_yofs);
		if (!count_editables(curline)) {
     		int r=seek_prev_editable_line(&cursor_sub, &cursor_ofs, &cursor_yofs);
			if (!r) {
     			r=-seek_next_editable_line(&cursor_sub, &cursor_ofs, &cursor_yofs);
				scrolltop=1;
			} else s++;
     		n+=r;
		} else s++;
     	cursor_y-=n;
		if (cursor_y<0) {
     		scroll_up(-cursor_y);
     		cursor_y=0;
     	}
		if (scrolltop) {
			int c=count_lines_to_top(currentsub, ofs, yofs);
          	if (c>0) {
				scroll_up(c);
               	cursor_y+=c;
			}
		}
		update_cursor_data();
	}
     return s;
}

int ht_uviewer::cursor_down(int n)
{
	int s=0;
	n=seek_next_line(&cursor_sub, &cursor_ofs, &cursor_yofs, n);
     if (n) {
     	bounds b;
          getbounds(&b);
		int scrolltop=0;
		cursor_sub->getline(curline, cursor_ofs, cursor_yofs);
		if (!count_editables(curline)) {
     		int r=seek_next_editable_line(&cursor_sub, &cursor_ofs, &cursor_yofs);
			if (!r) {
	     		r=-seek_prev_editable_line(&cursor_sub, &cursor_ofs, &cursor_yofs)+1;
				scrolltop=1;
			} else s++;
		    	n+=r;
		} else s++;
	     cursor_y+=n;
     	if (cursor_y>b.h-1) {
	     	scroll_down(cursor_y-b.h+1);
     		cursor_y=b.h-1;
	     }
		if (scrolltop) {
			int c=count_lines_to_bottom(currentsub, ofs, yofs)-b.h+1;
     	     if (c>0) {
				scroll_down(c);
     	          cursor_y-=c;
			}
		}
		update_cursor_data();
	}
     return s;
}

int ht_uviewer::cursor_left(int n)
{
     int o=n;
	while (n--) {
		if (cursor_x<=0) {
	          if (cursor_e>0) {
     	     	cursor_e--;
                    cursor_x=666;
			} else {
	               if (cursor_up(1)) cursor_end(); else break;
     	     }
     	} else cursor_x--;
     }
     update_cursor_data();
     return o-n;
}

int ht_uviewer::cursor_right(int n)
{
     int o=n;
	int c=count_editables_in_group(curline, cursor_group);
	while (n--) {
		char *e=get_editable(curline, cursor_e, cursor_group);
	     int s=1;
		if (edit) s= tag_get_vlen(e);
		if (cursor_x+1>=s) {
     	     if (cursor_e<c-1) {
	          	cursor_e++;
                    cursor_x=0;
			} else {
	               if (cursor_down(1)) cursor_home(); else break;
     	     }
    		} else cursor_x++;
     }
     update_cursor_data();
     return o-n;
}

/*
 *	CLASS ht_logviewer
 */

ht_logviewer::ht_logviewer(bounds *b, int _maxlinecount) : ht_viewer(b, "log", 0)
{
	maxlinecount=_maxlinecount;
     lines=new ht_string_collection();
     ofs=0;
}

ht_logviewer::~ht_logviewer()
{
	delete lines;
}

void ht_logviewer::deletefirstline()
{
	lines->del(0);
}

void ht_logviewer::draw()
{
	ht_viewer::draw();
     bounds b;
     getbounds(&b);
     int c=lines->count();
     for (int i=0; i<b.h; i++) {
     	if (i+ofs>=c) break;
          print(0, i, viewer_color_normal, lines->get(i+ofs));
     }
}

void ht_logviewer::insertline(char *line)
{
     if (lines->count()>=maxlinecount) deletefirstline();
     lines->add(line);
     dirtyview();
}

void ht_logviewer::sendmsg(htmsg *msg)
{
	ht_viewer::sendmsg(msg);
}

/*
 *	CLASS hts_sub
 */

hts_sub::hts_sub(htfile *_file)
{
     prev=0;
	next=0;
	file=_file;
}

hts_sub::~hts_sub()
{
}

ABSTRACT int hts_sub::getline(char *line, ADDR ofs, int yofs)
{
	abstract();
     return 0;
}

ABSTRACT void hts_sub::start_line_address(ADDR *ofs, int *yofs)
{
	abstract();
}

ABSTRACT void hts_sub::end_line_address(ADDR *ofs, int *yofs)
{
	abstract();
}

ABSTRACT int hts_sub::prev_line_address(ADDR *ofs, int *yofs, int n)
{
	abstract();
	return 0;
}

ABSTRACT int hts_sub::next_line_address(ADDR *ofs, int *yofs, int n)
{
	abstract();
	return 0;
}

STUB int hts_sub::jump(ADDR tofs, ADDR *ofs, int *yofs)
{
     return 0;
}

/*
 *	CLASS hts_hex
 */

hts_hex::hts_hex(htfile *file, int ofs, int size) : hts_sub(file)
{
	fofs=ofs;
     fsize=size;
}

hts_hex::~hts_hex()
{
}

int hts_hex::getline(char *line, ADDR ofs, int yofs)
{
	unsigned char buf[16];
	file->seek(ofs);
     int c=file->read(buf, 16);
     c=MIN(c, (int)(fofs+fsize-ofs));
     if (c<=0) return 0;
     char *l=line;
	l+=sprintf(l, "%08lx ", ofs);
     for (int i=0; i<16; i++) {
     	if (i<c) {
			l+=sprintf(l, HT_EDIT_BYTE_FORMAT " ", ofs+i, buf[i], tag_coloridx_unselected);
		} else {
			l+=sprintf(l, "   ");
          }
	}
	l+=sprintf(l, HT_GROUP_FORMAT);
     for (int i=0; i<16; i++) {
     	if (i<c) {
			l+=sprintf(l, HT_EDIT_CHAR_FORMAT, ofs+i, buf[i], tag_coloridx_unselected);
		} else {
			l+=sprintf(l, " ");
          }
	}
     return 1;
}

void hts_hex::start_line_address(ADDR *ofs, int *yofs)
{
	(*ofs)=fofs;
     (*yofs)=0;
}

void hts_hex::end_line_address(ADDR *ofs, int *yofs)
{
	(*ofs)=((fsize-1) & ~0xf)+fofs;
     (*yofs)=0;
}

int hts_hex::prev_line_address(ADDR *ofs, int *yofs, int n)
{
	int c=0;
     while (n--) {
          if (*ofs-16>*ofs) break;
          if (*ofs-16<fofs) break;
		(*ofs)-=16;
          c++;
/*        if (*ofs-16<fofs) break;
		(*ofs)-=16;
          c++;*/
     }
     return c;
}

int hts_hex::next_line_address(ADDR *ofs, int *yofs, int n)
{
	int c=0;
     while (n--) {
          if (*ofs+16>=fofs+fsize) break;
		(*ofs)+=16;
          c++;
     }
     return c;
}

int hts_hex::jump(ADDR tofs, ADDR *ofs, int *yofs)
{
	if ((tofs>=fofs) && (tofs<fofs+fsize)) {
     	*ofs=tofs;
          *yofs=0;
          return 1;
     }
     return 0;
}

/*
 *	CLASS hts_unasm
 */

hts_unasm::hts_unasm(htfile *file, ADDR ofs, int size) : hts_sub(file)
{
	fofs=ofs;
     fsize=size;
     unasm=new ht_unasm(32);
}

hts_unasm::~hts_unasm()
{
	delete unasm;
}

int hts_unasm::getline(char *line, ADDR ofs, int yofs)
{
	unsigned char buf[15];
	file->seek(ofs);
     int c=file->read(buf, 15);
     int g;
     c=MIN(c, (int)(fofs+fsize-ofs));
     if (c<=0) return 0;
     char *s=unasm->getstring(buf, c, 0, &g);
/*debug*/
	int c2=unasm->getsize(buf, c);
     assert(g==c2);
/*	if (g!=c2) {
     	printf("");
     	char *s2=unasm->getstring(buf, c, 0, &g);
		unasm->getsize(buf, c);
     }
/ **/
     char *l=line;
     l+=sprintf(l, "%08lx ", ofs);
     for (int i=0; i<15; i++) {
          if (i<g) {
     		l+=sprintf(l, HT_EDIT_BYTE_FORMAT, ofs+i, buf[i], tag_coloridx_unselected);
		} else {
     		l+=sprintf(l, "  ");
          }
     }
     l+=sprintf(l, " %s", s);
     return 1;
}

void hts_unasm::start_line_address(ADDR *ofs, int *yofs)
{
	(*ofs)=fofs;
     (*yofs)=0;
}

void hts_unasm::end_line_address(ADDR *ofs, int *yofs)
{
	(*ofs)=fofs+fsize;
     (*yofs)=0;
     prev_line_address(ofs, yofs, 1);
}

int hts_unasm::prev_line_address(ADDR *ofs, int *yofs, int n)
{
	unsigned char buf[15*50], *bufp=buf;
     int offsets[15*50];
     int *of=offsets;
     int r=n<6 ? 6*15 : n*15;
	ADDR o=*ofs-r;
	int c=r, d;
     int s;
     if (o<fofs) {
     	c-=fofs-o;
     	o=fofs;
     }
     if (o+c>fofs+fsize) {
     	c=fofs+fsize-o;
     }
    	if (c<=0) return 0;
	file->seek(o);
    	d=file->read(buf, c);
     do {
     	s=unasm->getsize(bufp, d);
          d-=s;
          *(of++)=o;
          o+=s;
          bufp+=s;
     } while (o<=*ofs);
     if (of-n-1<offsets) {
     	*ofs=*(offsets);
     	return of-offsets-1;
	} else {
     	*ofs=*(of-n-1);
     	return n;
	}
}

int hts_unasm::next_line_address(ADDR *ofs, int *yofs, int n)
{
	unsigned char buf[15];
	int c=0, s, z;
     while (n--) {
          z=MIN(15, fofs+fsize-*ofs);
		file->seek(*ofs);
     	z=file->read(buf, z);
          if (z<=0) break;
          s=unasm->getsize(buf, z);
          if (*ofs+s>=fofs+fsize) break;
          *ofs+=s;
          c++;
     }
     return c;
}

int hts_unasm::jump(ADDR tofs, ADDR *ofs, int *yofs)
{
	if ((tofs>=fofs) && (tofs<fofs+fsize)) {
     	*ofs=tofs;
          *yofs=0;
          return 1;
     }
     return 0;
}

/*
 *	CLASS hts_text
 */

hts_text::hts_text(htfile *file, const char *_text) : hts_sub(file)
{
	text=(char*)malloc(strlen(_text));
     strcpy(text, _text);
}

hts_text::~hts_text()
{
	if (text) free(text);
}

int hts_text::getline(char *line, ADDR ofs, int yofs)
{
	strcpy(line, text);
     return 1;
}

void hts_text::start_line_address(ADDR *ofs, int *yofs)
{
	*ofs=0;
     *yofs=0;
}

void hts_text::end_line_address(ADDR *ofs, int *yofs)
{
	*ofs=0;
     *yofs=0;
}

int hts_text::prev_line_address(ADDR *ofs, int *yofs, int n)
{
	return 0;
}

int hts_text::next_line_address(ADDR *ofs, int *yofs, int n)
{
    	return 0;
}

/*
 *	CLASS hts_mask
 */

hts_mask::hts_mask(htfile *file, int _start) : hts_sub(file)
{
     start=_start;
}

hts_mask::~hts_mask()
{
}

int hts_mask::getline(char *line, ADDR ofs, int yofs)
{
     char *l=temp;
     strcpy((char*)temp, masks.get(ofs));
     int c=count_editables(l);
     int g=count_groups(l);
     for (int j=0; j<g; j++) {
     	for (int i=0; i<c; i++) {
		     l=get_editable((char*)&temp, i, j);
               if ( tag_is_editable(l)) {
	               int o=tag_get_offset(l)+start;
     	          tag_set_offset(l, o);
          	     int v=0;
	               file->seek(o);
     	          file->read(&v, tag_get_size(l));
          	     tag_set_value(l, v);
			}
		}
	}
     strcpy(line, temp);
     return 1;
}

void hts_mask::start_line_address(ADDR *ofs, int *yofs)
{
	*ofs=0;
     *yofs=0;
}

void hts_mask::end_line_address(ADDR *ofs, int *yofs)
{
	*ofs=masks.count()-1;
     *yofs=0;
}

int hts_mask::prev_line_address(ADDR *ofs, int *yofs, int n)
{
	int r;
     if (*ofs<(dword)n) {
          r=*ofs;
     	*ofs=0;
     } else {
          r=n;
          *ofs-=n;
     }
     return r;
}

int hts_mask::next_line_address(ADDR *ofs, int *yofs, int n)
{
	int r=n;
     *ofs+=n;
     if ((int)*ofs>masks.count()-1) {
          r-=*ofs-masks.count()+1;
     	*ofs=masks.count()-1;
     }
     return r;
}

void hts_mask::add_mask(char *htstring)
{
	masks.add(htstring);
}

void hts_mask::add_mask_table(char **htstring)
{
     while (*htstring) add_mask(*(htstring++));
}

