#ifndef __HTOBJ_H__
#define __HTOBJ_H__

class ht_view;
class ht_group;

#include "common.h"
#include "htdebug.h"
#include "htcurses.h"
#include "htexpl.h"
#include "htfile.h"
#include "htdata.h"
#include "htunasm.h"
#include "viewers.h"

#define htm_nothing 		0
#define htm_draw 			1
#define htm_resize 			2
#define htm_keypressed 		3
#define htm_complete		4
#define htm_funcexec		5
#define htm_funcname		6
#define htm_funcname_return	7
#define htm_button_pressed	8

#define htmt_nothing		0
#define htmt_broadcast		1

struct htmsg {
	int msg;
     int type;
     union {
		int data_int;
		void *data_ptr;
          char *data_str;
     };
};

/*
 *	CLASS ht_view
 */

#define hto_ownbuffer		1
#define hto_browsable		2
#define hto_selectable		4

class ht_group;

class ht_view: public object {
protected:
    	genericdrawbuf *buf;
	int focused;
    	int view_is_dirty;
	bounds size;
     htpalette *palette;

	void cleanview();
	void dirtyview();
	int isviewdirty();
public:
	int options;
	char *desc;
	ht_group	*group;
	ht_view 	*prev, *next;

			ht_view(bounds *b, int options, char *desc);
	virtual 	~ht_view();
/* new */
			void box(int x, int y, int w, int h, int c);
			void clear();
			void clearmsg(htmsg *msg);
STUB	virtual 	void draw();
	virtual	int focus(ht_view *view);
			void getbounds(bounds *b);
	virtual	ht_view *getcurrent();
	virtual 	ht_view *getfirstchild();
     virtual	int getnumber();
	virtual 	void handlemsg(htmsg *msg);
     		void hidecursor();
			int lprint(int x, int y, int c, int l, char *text);
     		void move(int x, int y);
			int printf(int x, int y, int c, char *format, ...);
			int print(int x, int y, int c, char *text);
	virtual	void receivefocus();
			void redraw();
	virtual 	void releasefocus();
			void sendmsg(int msg, int data);
			void sendmsg(htmsg *msg);
			void setbounds(bounds *b);
               void setcursor(int x, int y);
               void setpalette(htpalette *palette);
};

/*
 *	CLASS ht_group
 */

class ht_group: public ht_view {
protected:
	ht_view 	*first, *current, *last;
public:
     void		*shared_data;

			ht_group(bounds *b, char *desc);
	virtual	~ht_group();
/* overwritten */
	virtual	int focus(ht_view *view);
	virtual	ht_view *getcurrent();
	virtual 	ht_view *getfirstchild();
STUB	virtual	void handlemsg(htmsg *msg);
			void putontop(ht_view *view);
	virtual	void redraw();
/* new */
			void insert(ht_view *view);
};

/*
 *	CLASS ht_xgroup
 */

class ht_xgroup: public ht_group {
public:
			ht_xgroup(bounds *b, char *desc);
	virtual	~ht_xgroup();
/* overwritten */
STUB	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_winborder
 */

class ht_winborder: public ht_view {
protected:
    	int number;
public:
			ht_winborder(bounds *b, char *desc, int number);
	virtual 	~ht_winborder();
/* overwritten */
STUB	virtual	void draw();
STUB	virtual	void handlemsg(htmsg *msg);
};

/*
 *	CLASS ht_window
 */

class ht_window: public ht_group {
protected:
	ht_winborder *border;
	int number;
public:
			ht_window(bounds *b, char *desc, int number);
	virtual 	~ht_window();
/* overwritten */
     virtual	int getnumber();
};

/*
 *	CLASS ht_viewer
 */

class ht_viewer: public ht_view {
protected:
	htfile *file;
public:
     void *shared_data;					/* data being shared with childs */

	ht_viewer(bounds *b, char *desc, htfile *file);
     virtual ~ht_viewer();
/* overwritten */
	STUB virtual void handlemsg(htmsg *msg);
/* new */
	STUB virtual void complete();
     STUB virtual int cursor_up(int n);
	STUB virtual int cursor_down(int n);
     STUB virtual int cursor_left(int n);
     STUB virtual int cursor_right(int n);
     STUB virtual char *func(int i, int getname);
     STUB virtual int jump(ADDR ofs);
     STUB virtual int jumpstr(char *str);
     STUB virtual void scroll_up(int n);
     STUB virtual void scroll_down(int n);
     STUB virtual void scroll_left(int n);
     STUB virtual void scroll_right(int n);
};

/*
 *	CLASS ht_uviewer
 */

class hts_sub; /* to resolve forward ref */

class ht_uviewer: public ht_viewer {
protected:
	hts_sub *firstsub, *currentsub, *lastsub, *cursor_sub;
	ADDR ofs;
	int yofs, xofs;
     ADDR cursor_eofs;
     ADDR cursor_ofs;
     int cursor_yofs;
     int cursor_e, cursor_x, cursor_y, cursor_group;
     char curline[1024];
     int edit;
/* new */
	void update_cursor_data();
	int seek_prev_line(hts_sub **sub, ADDR *ofs, int *yofs, int n);
	int seek_next_line(hts_sub **sub, ADDR *ofs, int *yofs, int n);
	int seek_prev_editable_line(hts_sub **sub, ADDR *ofs, int *yofs);
	int seek_next_editable_line(hts_sub **sub, ADDR *ofs, int *yofs);
	int count_lines_to_top(hts_sub *sub, ADDR ofs, int yofs);
	int count_lines_to_bottom(hts_sub *sub, ADDR ofs, int yofs);
	void apply_modifier(char *htstring);
	void cursor_home();
	void cursor_end();
	void cursor_next_group();
     int start_edit();
     int update_edit();
     int stop_edit();
     void qdraw();
	int editinput(int key);
     void popup_jump_dialog();
public:
		ht_uviewer(bounds *b, char *desc, htfile *file);
		virtual ~ht_uviewer();
/* overwritten */
     	virtual void draw();
		virtual void handlemsg(htmsg *msg);
     	virtual void complete();
     	virtual int cursor_up(int n);
	     virtual int cursor_down(int n);
     	virtual int cursor_left(int n);
	     virtual int cursor_right(int n);
		virtual char *func(int i, int getname);
		virtual int jump(ADDR ofs);
		virtual void scroll_up(int n);
		virtual void scroll_down(int n);
		virtual void scroll_left(int n);
		virtual void scroll_right(int n);
/* new */
	     virtual void insertsub(hts_sub *sub);
                  int htprint(int x, int y, int maxlen, int ofs, char *text);
};

/*
 *	CLASS ht_logviewer
 */

class ht_logviewer: public ht_viewer {
private:
	ht_string_collection *lines;
     int maxlinecount;
     int ofs;

     void deletefirstline();
public:
     ht_logviewer(bounds *b, int _maxlinecount);
     ~ht_logviewer();
/* overwritten */
     virtual void draw();
	virtual void sendmsg(htmsg *msg);
/* new */
     void insertline(char *_line);
};

/*
 *	CLASS hts_sub
 */

class hts_sub: public object {
protected:
	htfile *file;
public:
	hts_sub *prev, *next;

	hts_sub(htfile *file);
     virtual ~hts_sub();
/* new */
	ABSTRACT virtual int getline(char *line, ADDR ofs, int yofs);
     ABSTRACT virtual void start_line_address(ADDR *ofs, int *yofs);
     ABSTRACT virtual void end_line_address(ADDR *ofs, int *yofs);
     ABSTRACT virtual int prev_line_address(ADDR *ofs, int *yofs, int n);
     ABSTRACT virtual int next_line_address(ADDR *ofs, int *yofs, int n);
     STUB virtual int jump(ADDR tofs, ADDR *ofs, int *yofs);
};

/*
 *	CLASS hts_hex
 */

class hts_hex: public hts_sub {
private:
	ADDR fofs;
	dword fsize;
public:
	hts_hex(htfile *file, int ofs, int size);
	virtual ~hts_hex();
/* overwritten */
	virtual int getline(char *line, ADDR ofs, int yofs);
     virtual void start_line_address(ADDR *ofs, int *yofs);
     virtual void end_line_address(ADDR *ofs, int *yofs);
     virtual int prev_line_address(ADDR *ofs, int *yofs, int n);
     virtual int next_line_address(ADDR *ofs, int *yofs, int n);
    	virtual int jump(ADDR tofs, ADDR *ofs, int *yofs);
};

/*
 *	CLASS hts_unasm
 */

class hts_unasm: public hts_sub {
private:
	ADDR fofs;
	dword fsize;
     ht_unasm *unasm;
public:
	hts_unasm(htfile *file, ADDR ofs, int size);
	virtual ~hts_unasm();
/* overwritten */
	virtual int getline(char *line, ADDR ofs, int yofs);
     virtual void start_line_address(ADDR *ofs, int *yofs);
     virtual void end_line_address(ADDR *ofs, int *yofs);
     virtual int prev_line_address(ADDR *ofs, int *yofs, int n);
     virtual int next_line_address(ADDR *ofs, int *yofs, int n);
    	virtual int jump(ADDR tofs, ADDR *ofs, int *yofs);
};

/*
 *	CLASS hts_text
 */

class hts_text: public hts_sub {
private:
	char *text;
public:
	hts_text(htfile *file, const char *text);
	virtual ~hts_text();
/* overwritten */
	virtual int getline(char *line, ADDR ofs, int yofs);
     virtual void start_line_address(ADDR *ofs, int *yofs);
     virtual void end_line_address(ADDR *ofs, int *yofs);
     virtual int prev_line_address(ADDR *ofs, int *yofs, int n);
     virtual int next_line_address(ADDR *ofs, int *yofs, int n);
};

/*
 *	CLASS hts_mask
 */

class hts_mask: public hts_sub {
private:
	ht_string_collection masks;
     char temp[512];
     int start;
public:
	hts_mask(htfile *file, int start);
	virtual ~hts_mask();
/* overwritten */
	virtual int getline(char *line, ADDR ofs, int yofs);
     virtual void start_line_address(ADDR *ofs, int *yofs);
     virtual void end_line_address(ADDR *ofs, int *yofs);
     virtual int prev_line_address(ADDR *ofs, int *yofs, int n);
     virtual int next_line_address(ADDR *ofs, int *yofs, int n);
/* new */
	virtual void add_mask(char *htstring);
     virtual void add_mask_table(char **htstring);
};

#endif /* !__HTOBJ_H__ */
