/* 
 *	The HTE Editor
 *	htopcode.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htopcode.h"

/* 	Percent tokens in strings:
   	First char after '%':
        	A - direct address
        	C - reg of r/m picks control register
        	D - reg of r/m picks debug register
        	E - r/m picks operand
        	F - flags register
        	G - reg of r/m picks general register
        	I - immediate data (takes extended size, data size)
        	J - relative IP offset
        	M - r/m picks memory
        	O - no r/m, offset only
		P - reg of r/m picks mm register (mm0-mm7)
		Q - r/m picks mm operand (mm0-mm7/mem64)
        	R - mod of r/m picks register only
        	S - reg of r/m picks segment register
        	T - reg of r/m picks test register
        	X - DS:ESI
        	Y - ES:EDI
        	2 - prefix of two-unsigned char opcode
		3 - prefix of 3DNow! opcode
        	e - put in 'e' if use32 (second char is part of reg name)
		    put in 'w' for use16 or 'd' for use32 (second char is 'w')
		f - floating point (second char is esc value)
		g - do r/m group n (where n may be one of 0-9,A-Z)
        	p - prefix
        	s - size override (second char is a,o)
        	+ - make default signed
   	Second char after '%':
        	a - two words in memory (BOUND)
        	b - byte (8-bit integer)
        	c - "unsigned char or word" ??? (equal to 'v')
		d - dword (32-bit integer)
        	p - 32 or 48 bit pointer
		q - qword (64-bit integer)
        	s - six unsigned char pseudo-descriptor
        	v - word or dword
        	w - word (16-bit integer)
        	F - use floating regs in mod/rm
        	+ - always sign
        	- - sign if negative
        	1-8 - group number, esc value, etc
*/

char	*opmap[] = {
/* 0 */
  "add\t%Eb, %Gb", "add\t%Ev, %Gv", "add\t%Gb, %Eb", "add\t%Gv, %Ev",
  "add\tal, %I-bb", "add\t%eax, %I-vv", "push\tes", "pop\tes",
  "or\t%Eb, %Gb", "or\t%Ev, %Gv",	"or\t%Gb, %Eb", "or\t%Gv, %Ev",
  "or\tal, %Ibb", "or\t%eax, %Ivv", "push\tcs", "%2 ",
/* 1 */
  "adc\t%Eb, %Gb", "adc\t%Ev, %Gv", "adc\t%Gb, %Eb", "adc\t%Gv, %Ev",
  "adc\tal, %I-bb", "adc\t%eax, %I-vv", "push\tss", "pop\tss",
  "sbb\t%Eb, %Gb", "sbb\t%Ev, %Gv", "sbb\t%Gb, %Eb", "sbb\t%Gv, %Ev",
  "sbb\tal, %I-bb", "sbb\t%eax, %I-vv", "push\tds", "pop\tds",
/* 2 */
  "and\t%Eb, %Gb", "and\t%Ev, %Gv", "and\t%Gb, %Eb", "and\t%Gv, %Ev",
  "and\tal, %Ibb", "and\t%eax, %Ivv", "%pe", "daa",
  "sub\t%Eb, %Gb", "sub\t%Ev, %Gv", "sub\t%Gb, %Eb", "sub\t%Gv, %Ev",
  "sub\tal, %I-bb", "sub\t%eax, %I-vv", "%pc", "das",
/* 3 */
  "xor\t%Eb, %Gb", "xor\t%Ev, %Gv", "xor\t%Gb, %Eb", "xor\t%Gv, %Ev",
  "xor\tal, %Ibb", "xor\t%eax, %Ivv", "%ps", "aaa",
  "cmp\t%Eb, %Gb", "cmp\t%Ev, %Gv", "cmp\t%Gb, %Eb", "cmp\t%Gv, %Ev",
  "cmp\tal, %I-bb", "cmp\t%eax, %I-vv", "%pd", "aas",
/* 4 */
  "inc\t%eax", "inc\t%ecx", "inc\t%edx", "inc\t%ebx",
  "inc\t%esp", "inc\t%ebp", "inc\t%esi", "inc\t%edi",
  "dec\t%eax", "dec\t%ecx", "dec\t%edx", "dec\t%ebx",
  "dec\t%esp", "dec\t%ebp", "dec\t%esi", "dec\t%edi",
/* 5 */
  "push\t%eax",	"push\t%ecx", "push\t%edx", "push\t%ebx",
  "push\t%esp",	"push\t%ebp", "push\t%esi", "push\t%edi",
  "pop\t%eax", "pop\t%ecx", "pop\t%edx", "pop\t%ebx",
  "pop\t%esp", "pop\t%ebp", "pop\t%esi", "pop\t%edi",
/* 6 */
  "pusha", "popa", "bound\t%Gv, %Ma", "arpl\t%Ew, %Rw",
  "%pf", "%pg", "%so", "%sa",
  "push\t%I-vv", "imul\t%Gv, %Ev, %I-vv", "push\t%I-vb", "imul\t%Gv, %Ev, %I-vb",
  "insb\t%Yb,dx", "ins%ew\t%Yv,dx", "outsb\tdx, %Xb",	"outs%ew\tdx, %Xv",
/* 7 */
  "jo\t%Jb", "jno\t%Jb", "jc\t%Jb",  "jnc\t%Jb",
  "jz\t%Jb", "jnz\t%Jb", "jbe\t%Jb", "jnbe\t%Jb",
  "js\t%Jb", "jns\t%Jb", "jpe\t%Jb", "jpo\t%Jb",
  "jl\t%Jb", "jge\t%Jb", "jle\t%Jb", "jg\t%Jb",
/* 8 */
  "%g1\t%Eb, %Ibb", "%g1\t%Ev, %Ivv", 0, "%g1\t%Ev, %Ivb",
  "test\t%Eb, %Gb", "test\t%Ev, %Gv", "xchg\t%Eb, %Gb",	"xchg\t%Ev, %Gv",
  "mov\t%Eb, %Gb", "mov\t%Ev, %Gv", "mov\t%Gb, %Eb", "mov\t%Gv, %Ev",
  "mov\t%Ew, %Sw", "lea\t%Gv, %M ", "mov\t%Sw, %Ew", "pop\t%Ev",
/* 9 */
  "nop", "xchg\t%eax, %ecx", "xchg\t%eax, %edx",	"xchg\t%eax, %ebx",
  "xchg\t%eax, %esp",	"xchg\t%eax, %ebp", "xchg\t%eax, %esi", "xchg\t%eax, %edi",
  "cbw", "cwd", "call\t%Ap", "fwait",
  "push\t%eflags", "pop\t%eflags", "sahf", "lahf",
/* a */
  "mov\tal, %Ob", "mov\t%eax, %Ov", "mov\t%Ob,al", "mov\t%Ov, %eax",
  "movsb\t%Xb, %Yb", "movs%ew\t%Xv, %Yv", "cmpsb\t%Xb, %Yb",	"cmps%ew\t%Xv, %Yv",
  "test\tal, %Ibb", "test\t%eax, %Ivv", "stosb\t%Yb,al", "stos%ew\t%Yv, %eax",
  "lodsb\tal, %Xb ", "lods%ew\t%eax, %Xv", "scasb\tal, %Xb",	"scas%ew\t%eax, %Xv",
/* b */
  "mov\tal, %Ibb", "mov\tcl, %Ibb", "mov\tdl, %Ibb", "mov\tbl, %Ibb",
  "mov\tah, %Ibb", "mov\tch, %Ibb", "mov\tdh, %Ibb", "mov\tbh, %Ibb",
  "mov\t%eax, %I-vv",	"mov\t%ecx, %I-vv", "mov\t%edx, %I-vv", "mov\t%ebx, %I-vv",
  "mov\t%esp, %Ivv", "mov\t%ebp, %Ivv", "mov\t%esi, %I-vv", "mov\t%edi, %I-vv",
/* c */
  "%g2\t%Eb, %Ibb", "%g2\t%Ev, %Ibb", "ret\t%Iww", "ret",
  "les\t%Gv, %Mp", "lds\t%Gv, %Mp", "mov\t%Eb, %Ibb", "mov\t%Ev, %I-vv",
  "enter\t%Iww, %Ibb", "leave",	"retf\t%Iww", "retf",
  "int\t3", "int\t%Ibb", "into", "iret",
/* d */
  "%g2\t%Eb,1",	"%g2\t%Ev,1", "%g2\t%Eb,cl", "%g2\t%Ev,cl",
  "aam\t%Ibb", "aad\t%Ibb", "setalc", "xlat",
  "%f0", "%f1", "%f2", "%f3",
  "%f4", "%f5", "%f6", "%f7",
/* e */
  "loopne\t%Jb", "loope\t%Jb", "loop\t%Jb", "jcxz\t%Jb",
  "in\tal, %Ibb", "in\t%eax, %Ibb", "out\t%Ibb,al", "out\t%Ibb, %eax",
  "call\t%Jv", "jmp\t%Jv",	"jmp\t%Ap", "jmp\t%Jb",
  "in\tal, dx", "in\t%eax, dx", "out\tdx, al", "out\tdx, %eax",
/* f */
  "lock\t%p ", "smi", "repne\t%p ", "rep(e)\t%p ",
  "hlt", "cmc", "%g3", "%g0",
  "clc", "stc", "cli", "sti",
  "cld", "std", "%g4", "%g5"
};

char	*opmap_ext[] = {
/* 0 */
  "%g6", "%g7", "lar\t%Gv, %Ew", "lsl\t%Gv, %Ew", 0, 0, "clts",	0,
  "invd",	"wbinvd",	0, "ud2",	0, "prefetch\t%Eb", "femms", "%3 ",
/* 1 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 2 */
  "mov\t%Rd, %Cd", "mov\t%Rd, %Dd", "mov\t%Cd, %Rd", "mov\t%Dd, %Rd",
  "mov\t%Rd, %Td", 0,	"mov\t%Td, %Rd",	0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 3 */
  "wrmsr", "rdtsc",	"rdmsr", "rdpmc", "sysenter",	"sysexit", 0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 4 */
  "cmovo\t%Gv, %Ev", "cmovno\t%Gv, %Ev", "cmovc\t%Gv, %Ev", "cmovnc\t%Gv, %Ev",
  "cmovz\t%Gv, %Ev", "cmovnz\t%Gv, %Ev", "cmova\t%Gv, %Ev", "cmovna\t%Gv, %Ev",
  "cmovs\t%Gv, %Ev", "cmovns\t%Gv, %Ev", "cmovp\t%Gv, %Ev", "cmovnp\t%Gv, %Ev",
  "cmovl\t%Gv, %Ev", "cmovnl\t%Gv, %Ev", "cmovng\t%Gv, %Ev",	"cmovg\t%Gv, %Ev",
/* 5 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 6 */
  "punpcklbw\t%Pq, %Qd", "punpcklwd\t%Pq, %Qd", "punpckldq\t%Pq, %Qd",	"packsswb\t%Pq, %Qd",
  "pcmpgtb\t%Pq, %Qd", "pcmpgtw\t%Pq, %Qd",	"pcmpgtd\t%Pq, %Qd", "packuswb\t%Pq, %Qd",
  "punpckhbw\t%Pq, %Qd", "punpckhwd\t%Pq, %Qd", "punpckhdq\t%Pq, %Qd",	"packssdw\t%Pq, %Qd",
  0,	0, "movd\t%Pd, %Ed", "movq\t%Pq, %Qq",
/* 7 */
  0,	"%gA\t%Pq, %Ibb", "%gB\t%Pq, %Ibb", "%gC\t%Pq, %Ibb",
  "pcmpeqb\t%Pq, %Qq", "pcmpeqw\t%Pq, %Qq",	"pcmpeqd\t%Pq, %Qq", "emms",
  0,	0, 0, 0,
  0,	0, "movd\t%Ed, %Pd", "movq\t%Qq, %Pq",
/* 8 */
  "jo\t%Jv", "jno\t%Jv", "jc\t%Jv",  "jnc\t%Jv",
  "jz\t%Jv", "jnz\t%Jv", "jbe\t%Jv", "jnbe\t%Jv",
  "js\t%Jv", "jns\t%Jv", "jpe\t%Jv", "jpo\t%Jv",
  "jl\t%Jv", "jge\t%Jv", "jle\t%Jv", "jg\t%Jv",
/* 9 */
  "seto\t%Eb", "setno\t%Eb", "setc\t%Eb", "setnc\t%Eb",
  "setz\t%Eb", "setnz\t%Eb", "setbe\t%Eb",	"setnbe\t%Eb",
  "sets\t%Eb", "setns\t%Eb", "setpe\t%Eb",	"setpo\t%Eb",
  "setl\t%Eb", "setge\t%Eb", "setle\t%Eb",	"setg\t%Eb",
/* a */
  "push\tfs", "pop\tfs", "cpuid", "bt\t%Ev, %Gv",
  "shld\t%Ev, %Gv, %Ibb", "shld\t%Ev, %Gv,cl", 0,	0,
  "push\tgs", "pop\tgs", "rsm", "bts\t%Ev, %Gv",
  "shrd\t%Ev, %Gv, %Ibb", "shrd\t%Ev, %Gv,cl", 0,	"imul\t%Gv, %Ev",
/* b */
  "cmpxchg\t%Eb, %Gb", "cmpxchg\t%Ev, %Gv",	"lss\t%Mp", "btr\t%Ev, %Gv",
  "lfs\t%Mp", "lgs\t%Mp", "movzx\t%Gv, %Eb", "movzx\t%Gv, %Ew",
  0,	"ud2", "%g8\t%Ev, %Ibb", "btc\t%Ev, %Gv",
  "bsf\t%Gv, %Ev", "bsr\t%Gv, %Ev", "movsx\t%Gv, %Eb", "movsx\t%Gv, %Ew",
/* c */
  "xadd\t%Eb, %Gb", "xadd\t%Ev, %Gv", 0, 0,
  0,	0, 0, "%g9",
  "bswap\t%eax", "bswap\t%ecx",	"bswap\t%edx", "bswap\t%ebx",
  "bswap\t%esp", "bswap\t%ebp",	"bswap\t%esi", "bswap\t%edi",
/* d */
  0,	"psrlw\t%Pq, %Qq", "psrld\t%Pq, %Qq", "psrlq\t%Pq, %Qq",
  0,	"pmullw\t%Pq, %Qq", 0, 0,
  "psubusb\t%Pq, %Qq", "psubusw\t%Pq, %Qq",	0, "pand\t%Pq, %Qq",
  "paddusb\t%Pq, %Qq", "paddusw\t%Pq, %Qq",	0, "pandn\t%Pq, %Qq",
/* e */
  0,	"psraw\t%Pq, %Qq", "psrad\t%Pq, %Qq", 0,
  0,	"pmulhw\t%Pq, %Qq", 0, 0,
  "psubsb\t%Pq, %Qq",	"psubsw\t%Pq, %Qq", 0, "por\t%Pq, %Qq",
  "paddsb\t%Pq, %Qq",	"paddsw\t%Pq, %Qq", 0, "pxor\t%Pq, %Qq",
/* f */
  0,	"psllw\t%Pq, %Qq", "pslld\t%Pq, %Qq", "psllq\t%Pq, %Qq",
  0,	"pmuladdwd\t%Pq, %Qq", 0, 0,
  "psubb\t%Pq, %Qq", "psubw\t%Pq, %Qq", "psubq\t%Pq, %Qq", 0,
  "paddb\t%Pq, %Qq", "paddw\t%Pq, %Qq", "paddq\t%Pq, %Qq", 0
};

/* format of 3dnow instructions:
	0F 0F modR/M [sib] [displacement] 3dnow-suffix
   syntax of 3dnow instructions:
	mnemonic mm,mm/m64 (%Pq, %Qq)

examples:
   0F 0F CA B4			PFMUL mm1, mm2
   0F 0F 0B B4 		PFMUL mm1, [ebx]
   0F 0F 4B 0A B4 		PFMUL mm1, [ebx+10]
26 0F 0F 0B B4 		PFMUL mm1, es:[ebx]
   0F 0F 4C 83 0A B4 	PFMUL mm1, [ebx+eax*4+10]
*/

char	*opmap_ext2[] = {
/* 0 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, "pi2fd", 0, 0,
/* 1 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, "pf2id", 0, 0,
/* 2 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 3 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 4 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 5 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 6 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 7 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 8 */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* 9 */
  "pfcmpge", 0, 0, 0,
  "pfmin", 0, "pfrcp", "pfrsqrt",
  0,	0, "pfsub", 0,
  0,	0, "pfadd", 0,
/* a */
  "pfcmpgt", 0, 0, 0,
  "pfmax", 0, "pfrcpit1", "pfrsqit1",
  0,	0, "pfsubr", 0,
  0,	0, "pfacc", 0,
/* b */
  "pfcmpeq", 0, 0, 0,
  "pfmul", 0, "pfrcpit2", "pmulhrw",
  0,	0, 0, 0,
  0,	0, 0, "pavgusb",
/* c */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* d */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* e */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0,
/* f */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	0, 0, 0, 0, 0,	0, 0
};

char	*groups[][8] =	{	 /* group 0 is group 3 for %Ev set */
/* 0 */
  { "test\t%Ev, %Ivv", "test\t%Ev, %Ivv,", "not\t%Ev",	"neg\t%Ev",
    "mul\t%eax, %Ev",	"imul\t%eax, %Ev", "div\t%eax, %Ev", "idiv\t%eax, %Ev" },
/* 1 */
  { "add%+-", "or",	"adc%+-",	"sbb%+-",	"and", "sub%+-", "xor", "cmp%+-" },
/* 2 */
  { "rol", "ror", "rcl",	"rcr", "shl", "shr", "shl", "sar" },
/* 3 */
  { "test\t%Eb, %Ibb", "test\t%Eb, %Ibb,", "not\t%Eb",	"neg\t%Eb",
    "mul\tal, %Eb", "imul\tal, %Eb", "div\tal, %Eb", "idiv\tal, %Eb" },
/* 4 */
  { "inc\t%Eb",	"dec\t%Eb", 0, 0, 0,	0, 0, 0 },
/* 5 */
  { "inc\t%Ev",	"dec\t%Ev", "call\t%Ev", "call\t%Ep",
    "jmp\t%Ev",	"jmp\t%Ep", "push\t%Ev", 0	},
/* 6 */
  { "sldt\t%Ew", "str\t%Ew", "lldt\t%Ew", "ltr\t%Ew",
    "verr\t%Ew", "verw\t%Ew", 0, 0 },
/* 7 */
  { "sgdt\t%Ms", "sidt\t%Ms", "lgdt\t%Ms",	"lidt\t%Ms",
    "smsw\t%Ew", 0, "lmsw\t%Ew", 0 },
/* 8 */
  { 0, 0,	0, 0, "bt", "bts", "btr", "btc" },
/* 9 */
  { 0, "cmpxchg8b\t%Mq", 0, 0,	0, 0, 0, 0 },
/* A */
  { 0, 0,	"psrlw", 0, "psraw", 0, "psllw", 0	},
/* B (same as A, but d instead of w, non-intel group...) */
  { 0, 0,	"psrld", 0, "psrad", 0, "pslld", 0	},
/* C (same as A, but q instead of w, non-intel group...) */
  { 0, 0,	"psrlq", 0, "psraq", 0, "psllq", 0	}
  };

/* zero here means invalid.  If first entry starts with '*', use st(i) */
/* no assumed %EFs here.  Indexed by rm(modrm()) */
char	*f0[] = {0, 0,	0, 0, 0, 0, 0,	0};
char	*fop_9[]	= { "*fxch\tst, %GF" };
char	*fop_10[]	= { "fnop", 0,	0, 0, 0, 0, 0,	0 };
char	*fop_12[]	= { "fchs", "fabs",	0, 0, "ftst", "fxam", 0,	0 };
char	*fop_13[]	= { "fld1", "fldl2t", "fldl2e", "fldpi",
			    "fldlg2", "fldln2", "fldz", 0 };
char	*fop_14[]	= { "f2xm1", "fyl2x", "fptan", "fpatan",
			    "fxtract",	"fprem1",	"fdecstp", "fincstp" };
char	*fop_15[]	= { "fprem", "fyl2xp1", "fsqrt", "fsincos",
			    "frndint",	"fscale",	"fsin", "fcos"	};
char	*fop_21[]	= { 0, "fucompp", 0, 0, 0, 0,	0, 0	};
char	*fop_28[]	= { 0, 0,	"fclex", "finit", 0, 0, 0, 0 };
char	*fop_32[]	= { "*fadd\t%GF, st" };
char	*fop_33[]	= { "*fmul\t%GF, st" };
char	*fop_36[]	= { "*fsubr\t%GF, st"	};
char	*fop_37[]	= { "*fsub\t%GF,  st" };
char	*fop_38[]	= { "*fdivr\t%GF, st"	};
char	*fop_39[]	= { "*fdiv\t%GF, st" };
char	*fop_40[]	= { "*ffree\t%GF" };
char	*fop_42[]	= { "*fst\t%GF"	};
char	*fop_43[]	= { "*fstp\t%GF" };
char	*fop_44[]	= { "*fucom\t%GF" };
char	*fop_45[]	= { "*fucomp\t%GF" };
char	*fop_48[]	= { "*faddp\t%GF, st" };
char	*fop_49[]	= { "*fmulp\t%GF, st" };
char	*fop_51[]	= { 0, "fcompp", 0,	0, 0, 0, 0, 0 };
char	*fop_52[]	= { "*fsubrp\t%GF, st" };
char	*fop_53[]	= { "*fsubp\t%GF, st" };
char	*fop_54[]	= { "*fdivrp\t%GF, st" };
char	*fop_55[]	= { "*fdivp\t%GF, st" };
char	*fop_60[]	= { "fstsw\tax", 0, 0, 0, 0, 0, 0, 0 };

char	**fspecial[] =	{ /* 0=use st(i), 1=undefined 0 in fop_* means undefined */
  0,	0, 0, 0, 0, 0,	0, 0,
  0,	fop_9, fop_10,	0, fop_12, fop_13, fop_14, fop_15,
  f0, f0,	f0, f0, f0, fop_21,	f0, f0,
  f0, f0,	f0, f0, fop_28, f0,	f0, f0,
  fop_32,	fop_33, f0, f0, fop_36, fop_37, fop_38,	fop_39,
  fop_40,	f0, fop_42, fop_43,	fop_44, fop_45, f0,	f0,
  fop_48,	fop_49, f0, fop_51,	fop_52, fop_53, fop_54, fop_55,
  f0, f0,	f0, f0, fop_60, f0,	f0, f0,
  };

char	*floatops[] = { /* assumed "%EF" at end of each.  mod != 3 only */
/*00*/ "fadd",	"fmul", "fcom", "fcomp",
	  "fsub",	"fsubr", "fdiv", "fdivr",
/*08*/ "fld", 0, "fst", "fstp",
	  "fldenv", "fldcw", "fstenv", "fstcw",
/*16*/ "fiadd", "fimul",	"ficomw",	"ficompw",
	  "fisub", "fisubr", "fidiv",	"fidivr",
/*24*/ "fild",	0, "fist", "fistp",
	  "frstor", "fldt",	0, "fstpt",
/*32*/ "faddq", "fmulq",	"fcomq", "fcompq",
	  "fsubq", "fsubrq", "fdivq",	"fdivrq",
/*40*/ "fldq",	0, "fstq", "fstpq",
	  0,	0, "fsave", "fstsww",
/*48*/ "fiaddw", "fimulw", "ficomw", "ficompw",
	  "fisubw", "fisubrw", "fidivw", "fidivr",
/*56*/ "fildw", 0, "fistw", "fistpw",
	  "fbldt", "fildq",	"fbstpt",	"fistpq"
  };

