/* 
 *	The HTE Editor
 *	htpedimp.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htapp.h"
#include "htfile.h"
#include "htnewexe.h"
#include "htpe.h"
#include "htpedimp.h"
#include "htstring.h"
#include "viewers.h"

#include <stdlib.h>

int htpedelayimports_detect(htfile *file)
{
	int h=get_newexe_header_ofs(file);
     dword pe=0;
     file->seek(h);
     file->read(&pe, 2);
	if (pe==PE_MAGIC) {
	     int e, f;
	     file->seek(h+0x78+13*8);
     	file->read(&e, 4);
     	file->read(&f, 4);
		return e|f;
	}
     return 0;
}

ht_view *htpedelayimports_init(bounds *b, htfile *file, ht_group *group)
{
     ht_pe_shared_data *pe_shared=(ht_pe_shared_data *)((ht_pe*)group)->shared_data;
     dword base=pe_shared->pe32header.image_base;
     int h=get_newexe_header_ofs(file);
	char iline[1024];	/* FIXME: possible buffer overflow */
	ht_uviewer *v=new ht_uviewer(b, "pe/delay-imports", file);
     hts_mask *m=new hts_mask(file, h);

/* get delay import directory offset */
	/* 1. get import directory rva */
     dword iofs, irva=pe_shared->pe32header.directory[PE_DIRECTORY_ENTRY_DELAY_IMPORT].address;
     /* 2. transform it into an offset */
     if (!pe_rva_to_ofs(&pe_shared->sections, irva, &iofs)) goto pe_read_error;
	LOG("PE: reading delay-import directory at offset %08x, rva %08x, size %08x...", iofs, irva, pe_shared->pe32header.directory[PE_DIRECTORY_ENTRY_DELAY_IMPORT].size);
/*** read delay import directory ***/
	PE_DELAY_IMPORT_DESCRIPTOR dimport;
     int dll_index;
     dll_index=0;
     dword f;
     f=0;

     while (1) {
    		file->seek(iofs);
		file->read(&dimport, sizeof dimport);
          if (!dimport.name) break;
          dword dllname_ofs;
		if (!pe_rva_to_ofs(&pe_shared->sections, dimport.name-base, &dllname_ofs)) goto pe_read_error;
          file->seek(dllname_ofs);
          char *dllname=getstrz(file);
          ht_pe_dimport_library *lib=new ht_pe_dimport_library(dllname);
          pe_shared->dimports.libs.add(lib);

          dword ntofs=dimport.delay_int-base, nva, nofs;
          dword atofs=dimport.delay_iat-base, ava;
          while (1) {
               ht_pe_dimport_function *func;
          	file->seek(ntofs);
               file->read(&nva, 4);
               if (!nva) break;
          	file->seek(atofs);
               file->read(&ava, 4);
               if (nva & 0x80000000) {
               	/* import by ordinal */
				sprintf(iline, "%s: "_SEL("%08lx", "%02lx")"%04lx", dllname, ntofs, (dword)4, nva&0xffff);
				m->add_mask(iline);
               	func=new ht_pe_dimport_function(dll_index, ava, nva&0xffff);
               } else {
               	/* import by name */
				if (!pe_rva_to_ofs(&pe_shared->sections, nva-base, &nofs)) goto pe_read_error;
	               int hint=0;
     	     	file->seek(nofs);
          	     file->read(&hint, 2);
          		char *name=getstrz(file);
				sprintf(iline, "%s: "_SEL("%08lx", "%02lx")"%s (hint=%04x)", dllname, f, strlen(name)+12, name, hint);
				m->add_mask(iline);
               	func=new ht_pe_dimport_function(dll_index, ava, name, hint);
               	delete name;
			}
               pe_shared->dimports.funcs.add(func);
               ntofs+=4;
               atofs+=4;
               f++;
          }

          delete dllname;
          iofs+=sizeof dimport;
          dll_index++;
     }

//     m->sort_masks();
     m->add_mask(_SEL("ffffffff", "05")"<END>");
     v->insertsub(m);

	return v;
pe_read_error:
     assert(0);
	return 0;
}

viewer_int htpedelayimports_int = {
	htpedelayimports_detect,
	htpedelayimports_init
};

/*
 *	class ht_pe_dimport_library
 */

ht_pe_dimport_library::ht_pe_dimport_library()
{
}

ht_pe_dimport_library::ht_pe_dimport_library(char *name)
{
	this->name=htstrdup(name);
}

ht_pe_dimport_library::~ht_pe_dimport_library()
{
}

/*
 *	class ht_pe_dimport_function
 */

ht_pe_dimport_function::ht_pe_dimport_function()
{
	dll=0;
     name.name=0;
     name.hint=0;
     ordinal=0;
     function_address=0;
}

ht_pe_dimport_function::ht_pe_dimport_function(int dll, int function_address, int ordinal)
{
	this->dll=dll;
     this->ordinal=ordinal;
     this->function_address=function_address;
     byname=0;
}

ht_pe_dimport_function::ht_pe_dimport_function(int dll, int function_address, char *name, int hint)
{
	this->dll=dll;
     this->name.name=htstrdup(name);
     this->name.hint=hint;
     this->function_address=function_address;
     byname=1;
}

ht_pe_dimport_function::~ht_pe_dimport_function()
{
     if ((byname) && (name.name)) delete name.name;
}
