/* 
 *	The HTE Editor
 *	htstring.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "global.h"
#include "htdebug.h"
#include "htstring.h"

#include <stdlib.h>
#include <string.h>

int hexb(char *s)
{
	char b=*s;
     b-='0';
     if (b>9) b-='a'-'0'-10;
	char c=*(s+1);
     c-='0';
     if (c>9) c-='a'-'0'-10;
     return (b<<4)+c;
}

int hexw(char *s)
{
	return (hexb(s)<<8) | hexb(s+2);
}

int hexd(char *s)
{
	return (hexw(s)<<16) | hexw(s+4);
}

#define optimized_findtag

char *findtag(char *htstring)
{
#ifdef optimized_findtag
     int *s=(int*)htstring;
     int e;
     while (*s) {
     	e=*(s++);
          for (int i=0; i<4; i++) {
               if ((e&0xff)=='\e') return (char*)htstring;
               if ((e&0xff)==0) return 0;
			e=e>>8;
               htstring++;
          }
	}
     return 0;
#else
	return strchr(htstring, '\e');
#endif
}

int tag_get_len(char *htstring)
{
	assert(htstring[0]=='\e');
	switch (htstring[1]) {
         	case HT_EDIT_BYTE:
          	return HT_EDIT_BYTE_LEN;
         	case HT_EDIT_WORD:
          	return HT_EDIT_WORD_LEN;
         	case HT_EDIT_DWORD:
          	return HT_EDIT_DWORD_LEN;
         	case HT_EDIT_TIME:
          	return HT_EDIT_TIME_LEN;
		case HT_EDIT_CHAR:
           	return HT_EDIT_CHAR_LEN;
         	case HT_EXPL:
	         	return HT_EXPL_LEN;
		case HT_SEL:
          	return HT_SEL_LEN;
		case HT_QUOTE: {
               int z;
               if (!sscanf(htstring, HT_QUOTE_FORMAT, &z)) assert(0);
          	return HT_QUOTE_LEN(z);
		}
		case HT_GROUP:
          	return HT_GROUP_LEN;
	}
	assert(0);
	return -1;
}

int tag_get_vlen(char *htstring)
{
	assert(htstring[0]=='\e');
	switch (htstring[1]) {
         	case HT_EDIT_BYTE:
          	return HT_EDIT_BYTE_VLEN;
         	case HT_EDIT_WORD:
          	return HT_EDIT_WORD_VLEN;
         	case HT_EDIT_DWORD:
          	return HT_EDIT_DWORD_VLEN;
         	case HT_EDIT_TIME:
          	return HT_EDIT_TIME_VLEN;
		case HT_EDIT_CHAR:
          	return HT_EDIT_CHAR_VLEN;
		case HT_SEL: {
               int l=hexb(htstring+10);
          	return HT_SEL_VLEN(l);
		}
		default:
          	return 0;
	}
}

int tag_get_size(char *htstring)
{
	assert(htstring[0]=='\e');
	switch (htstring[1]) {
         	case HT_EDIT_BYTE:
          	return HT_EDIT_BYTE_SIZE;
         	case HT_EDIT_WORD:
          	return HT_EDIT_WORD_SIZE;
         	case HT_EDIT_DWORD:
          	return HT_EDIT_DWORD_SIZE;
         	case HT_EDIT_TIME:
          	return HT_EDIT_TIME_SIZE;
		case HT_EDIT_CHAR:
          	return HT_EDIT_CHAR_SIZE;
		case HT_SEL:
			return HT_SEL_SIZE;
		default:
          	return 0;
	}
}

dword tag_get_offset(char *htstring)
{
	assert(htstring[0]=='\e');
	switch (htstring[1]) {
         	case HT_EDIT_BYTE:
         	case HT_EDIT_WORD:
         	case HT_EDIT_DWORD:
         	case HT_EDIT_TIME:
         	case HT_EDIT_CHAR:
         	case HT_SEL:
               return hexd(htstring+2);
	}
	assert(0);
	return 0;
}

int tag_get_value(char *htstring)
{
	assert(htstring[0]=='\e');
	switch (htstring[1]) {
         	case HT_EDIT_BYTE:
         	case HT_EDIT_WORD:
         	case HT_EDIT_DWORD:
         	case HT_EDIT_TIME:
         	case HT_EDIT_CHAR:
               return hexb(htstring+10);
	}
     assert(0);
	return -1;
}

void tag_set_color(char *htstring, int color)
{
	assert(htstring[0]=='\e');
	switch (htstring[1]) {
         	case HT_EDIT_BYTE: {
               char buf[3];
               sprintf(buf, "%02x", color);
               memmove(htstring+12, buf, 2);
               break;
		}
         	case HT_EDIT_WORD: {
               char buf[3];
               sprintf(buf, "%02x", color);
               memmove(htstring+14, buf, 2);
               break;
		}
         	case HT_EDIT_DWORD: {
               char buf[3];
               sprintf(buf, "%02x", color);
               memmove(htstring+18, buf, 2);
               break;
		}
         	case HT_EDIT_TIME: {
               char buf[3];
               sprintf(buf, "%02x", color);
               memmove(htstring+18, buf, 2);
               break;
		}
		case HT_EDIT_CHAR: {
               char buf[3];
               sprintf(buf, "%02x", color);
               memmove(htstring+12, buf, 2);
               break;
		}
		case HT_SEL: {
               char buf[3];
               sprintf(buf, "%02x", color);
               memmove(htstring+12, buf, 2);
               break;
		}
	}
}

void tag_set_offset(char *htstring, dword offset)
{
	assert(htstring[0]=='\e');
	switch (htstring[1]) {
		case HT_EDIT_CHAR:
         	case HT_EDIT_BYTE:
         	case HT_EDIT_WORD:
         	case HT_EDIT_DWORD:
         	case HT_EDIT_TIME:
         	case HT_SEL: {
               char buf[9];
               sprintf(buf, "%08lx", offset);
               memmove(htstring+2, buf, 8);
               break;
		}
	}
}

void tag_set_value(char *htstring, int value)
{
	assert(htstring[0]=='\e');
	switch (htstring[1]) {
         	case HT_EDIT_BYTE: {
               char buf[3];
               sprintf(buf, "%02x", value);
               memmove(htstring+10, buf, 2);
               break;
		}
         	case HT_EDIT_WORD: {
			char buf[5];
			sprintf(buf, "%04x", value);
			memmove(htstring+10, buf, 4);
               break;
		}
         	case HT_EDIT_DWORD: {
               char buf[9];
               sprintf(buf, "%08x", value);
               memmove(htstring+10, buf, 8);
               break;
		}
         	case HT_EDIT_TIME: {
               char buf[9];
               sprintf(buf, "%08x", value);
               memmove(htstring+10, buf, 8);
               break;
		}
		case HT_EDIT_CHAR: {
               char buf[3];
               sprintf(buf, "%02x", value);
               memmove(htstring+10, buf, 2);
               break;
		}
	}
}

void tag_set_desc(char *htstring, void *desc)
{
	assert(htstring[0]=='\e');
	switch (htstring[1]) {
         	case HT_EDIT_BYTE: {
               char buf[9];
               sprintf(buf, "%08x", (int)desc);
               memmove(htstring+14, buf, 8);
               break;
		}
         	case HT_EDIT_WORD: {
			char buf[9];
			sprintf(buf, "%08x", (int)desc);
			memmove(htstring+16, buf, 8);
               break;
		}
         	case HT_EDIT_DWORD: {
               char buf[9];
               sprintf(buf, "%08x", (int)desc);
               memmove(htstring+20, buf, 8);
               break;
		}
         	case HT_EDIT_TIME: {
               char buf[9];
               sprintf(buf, "%08x", (int)desc);
               memmove(htstring+20, buf, 8);
               break;
		}
	}
}

int tag_is_editable(char *htstring)
{
	assert(htstring[0]=='\e');
	switch (htstring[1]) {
		case HT_EDIT_CHAR:
         	case HT_EDIT_BYTE:
         	case HT_EDIT_WORD:
         	case HT_EDIT_DWORD:
         	case HT_EDIT_TIME:
          	return 1;
         	case HT_SEL:
          	return 0;
	}
     assert(0);
     return -1;
}

int htstrlen(char *htstring)
{
	int c=0;
     while (*htstring) {
     	if (htstring[0]=='\e') {
          	int s=tag_get_len(htstring);
          	c+=tag_get_vlen(htstring);
               htstring+=s;
		} else {
			c++;
			htstring++;
		}
	}
     return c;
}

int count_editables_in_group(char *htstring, int group)
{
	int c=0;
     htstring=get_group(htstring, group);
     while ((htstring=findtag(htstring))) {
    		switch (htstring[1]) {
              	case HT_EDIT_BYTE:
                   	c++;
				htstring+=HT_EDIT_BYTE_LEN;
				break;
              	case HT_EDIT_WORD:
                   	c++;
				htstring+=HT_EDIT_WORD_LEN;
				break;
              	case HT_EDIT_DWORD:
                   	c++;
				htstring+=HT_EDIT_DWORD_LEN;
				break;
              	case HT_EDIT_TIME:
                   	c++;
				htstring+=HT_EDIT_TIME_LEN;
				break;
			case HT_EDIT_CHAR:
				c++;
				htstring+=HT_EDIT_CHAR_LEN;
				break;
			case HT_SEL:
				c++;
				htstring+=HT_SEL_LEN;
				break;
			case HT_GROUP:
               	return c;
			default:
          		htstring+=tag_get_len(htstring);
               	break;
          }
     }
     return c;
}

int count_editables(char *htstring)
{
	int c=0;
     while ((htstring=findtag(htstring))) {
    		switch (htstring[1]) {
              	case HT_EDIT_BYTE:
                   	c++;
				htstring+=HT_EDIT_BYTE_LEN;
				break;
              	case HT_EDIT_WORD:
                   	c++;
				htstring+=HT_EDIT_WORD_LEN;
				break;
              	case HT_EDIT_DWORD:
                   	c++;
				htstring+=HT_EDIT_DWORD_LEN;
				break;
              	case HT_EDIT_TIME:
                   	c++;
				htstring+=HT_EDIT_TIME_LEN;
				break;
			case HT_EDIT_CHAR:
				c++;
				htstring+=HT_EDIT_CHAR_LEN;
				break;
			case HT_SEL:
				c++;
				htstring+=HT_SEL_LEN;
				break;
			default:
          		htstring+=tag_get_len(htstring);
               	break;
          }
     }
     return c;
}

int count_groups(char *htstring)
{
	int c=1;
     while ((htstring=findtag(htstring))) {
    		switch (htstring[1]) {
              	case HT_GROUP:
               	c++;
				htstring+=HT_GROUP_LEN;
				break;
			default:
          		htstring+=tag_get_len(htstring);
               	break;
          }
     }
     return c;
}

char *get_editable(char *htstring, int n, int group)
{
	char *r=0;
     if (group) htstring=get_group(htstring, group);
     n++;
     while ((n) && (htstring=findtag(htstring))) {
    		switch (htstring[1]) {
              	case HT_EDIT_BYTE:
                   	n--;
                    r=htstring;
				htstring+=HT_EDIT_BYTE_LEN;
				break;
              	case HT_EDIT_WORD:
                   	n--;
                    r=htstring;
				htstring+=HT_EDIT_WORD_LEN;
				break;
              	case HT_EDIT_DWORD:
                   	n--;
                    r=htstring;
				htstring+=HT_EDIT_DWORD_LEN;
				break;
              	case HT_EDIT_TIME:
                   	n--;
                    r=htstring;
				htstring+=HT_EDIT_TIME_LEN;
				break;
			case HT_EDIT_CHAR:
				n--;
                    r=htstring;
				htstring+=HT_EDIT_CHAR_LEN;
				break;
			case HT_SEL:
				n--;
                    r=htstring;
				htstring+=HT_SEL_LEN;
				break;
			case HT_GROUP:
                    return (char*)r;
			default:
          		htstring+=tag_get_len(htstring);
               	break;
		}
     }
     return (char*)r;
}

char *get_group(char *htstring, int group)
{
	char *r=htstring;
     while ((group) && (htstring=findtag(htstring))) {
    		switch (htstring[1]) {
              	case HT_GROUP:
                   	group--;
				htstring+=HT_GROUP_LEN;
                    r=htstring;
				break;
			default:
          		htstring+=tag_get_len(htstring);
               	break;
		}
     }
     return (char*)r;
}

char *htstrdup(char *string)
{
     char *s=(char*)malloc(strlen(string)+1);
     strcpy(s, string);
     return s;
}

void htstriptags(char *s)
{
	while ((s=findtag(s))) {
     	int l=strlen(s);
     	switch (s[1]) {
          	case HT_EDIT_BYTE:
                    memmove(s, s+10, l-10+1);
                    memmove(s+2, s+4, l-2+1);
               	break;
          	case HT_EDIT_WORD:
                    memmove(s, s+10, l-10+1);
                    memmove(s+2, s+6, l-4+1);
               	break;
          	case HT_EDIT_DWORD:
                    memmove(s, s+10, l-10+1);
                    memmove(s+2, s+10, l-8+1);
               	break;
          	case HT_SEL:
                    memmove(s, s+HT_SEL_LEN, l-HT_SEL_LEN+1);
               	break;
			default:
               	assert(0);
                    break;
          }
     }
}

