#ifndef __HTSTRING_H__
#define __HTSTRING_H__

#include "config.h"

#include "htlayout.h"
#include "global.h"

/* Editable: byte ("\e\1(%08x)(%02x)(%02x)" $1=file offset, $2=value, $3=color */
#define HT_EDIT_BYTE 1
#define HT_EDIT_BYTE_CH "\e\1"
#define HT_EDIT_BYTE_LEN (2+8+2+2)
#define HT_EDIT_BYTE_VLEN (2)
#define HT_EDIT_BYTE_SIZE (1)
#define HT_EDIT_BYTE_FORMAT HT_EDIT_BYTE_CH"%08lx%02x%02x"

/* Editable: word ("\e\2(%08x)(%04x)(%02x)" $1=file offset, $2=value, $3=color */
#define HT_EDIT_WORD 2
#define HT_EDIT_WORD_CH "\e\2"
#define HT_EDIT_WORD_LEN (2+8+4+2)
#define HT_EDIT_WORD_VLEN (4)
#define HT_EDIT_WORD_SIZE (2)
#define HT_EDIT_WORD_FORMAT HT_EDIT_WORD_CH"%08lx%04x%02x"

/* Editable: dword ("\e\3(%08x)(%08x)(%02x)" $1=file offset, $2=value, $3=color */
#define HT_EDIT_DWORD 3
#define HT_EDIT_DWORD_CH "\e\3"
#define HT_EDIT_DWORD_LEN (2+8+8+2)
#define HT_EDIT_DWORD_VLEN (8)
#define HT_EDIT_DWORD_SIZE (4)
#define HT_EDIT_DWORD_FORMAT HT_EDIT_DWORD_CH"%08lx%08x%02x"

/* Editable: time ("\e\4(%08x)(%08x)(%02x)" $1=file offset, $2=value, $3=color */
#define HT_EDIT_TIME 4
#define HT_EDIT_TIME_CH "\e\4"
#define HT_EDIT_TIME_LEN (2+8+8+2)
#define HT_EDIT_TIME_VLEN (19)
#define HT_EDIT_TIME_SIZE (4)
#define HT_EDIT_TIME_FORMAT HT_EDIT_TIME_CH"%08lx%08x%02x"

/* Editable: char ("\e\5(%08x)(%02x)(%02x)" $1=file offset, $2=value, $3=color) */
#define HT_EDIT_CHAR 5
#define HT_EDIT_CHAR_CH "\e\5"
#define HT_EDIT_CHAR_LEN (2+8+2+2)
#define HT_EDIT_CHAR_VLEN (1)
#define HT_EDIT_CHAR_SIZE (1)
#define HT_EDIT_CHAR_FORMAT HT_EDIT_CHAR_CH"%08lx%02x%02x"

/* Selectable ("\e\6(%08x)(%02x)(%02x)(%08x)" $1=preferrably unique id, $2=length, $3=color, $4=configurable id) */
#define HT_SEL 6
#define HT_SEL_CH "\e\6"
#define HT_SEL_LEN (2+8+2+2+8)
#define HT_SEL_VLEN(n) (n)
#define HT_SEL_SIZE (0)
#define HT_SEL_FORMAT HT_SEL_CH"%08x%02x%02x%08x"

/* Explanation ("\e\7(%08x)" $1=unique expl_table id) */
#define HT_EXPL 7
#define HT_EXPL_CH "\e\7"
#define HT_EXPL_LEN (2+8)
#define HT_EXPL_FORMAT HT_EXPL_CH"%08x"

/* Quote (do not interpret) "\e\x8(%02x)" $1=number of chars to quote */
#define HT_QUOTE 8
#define HT_QUOTE_CH "\e\x8"
#define HT_QUOTE_LEN(n) (2+(n))
#define HT_QUOTE_FORMAT HT_QUOTE_CH"%02x"

/* Group separator ("\e\x9") */
#define HT_GROUP 9
#define HT_GROUP_CH "\e\x9"
#define HT_GROUP_LEN (2)
#define HT_GROUP_FORMAT HT_GROUP_CH

#define tag_coloridx_normal			1
#define tag_coloridx_normal_str		"01"
#define tag_coloridx_selected			2
#define tag_coloridx_selected_str		"02"
#define tag_coloridx_unselected		3
#define tag_coloridx_unselected_str	"03"
#define tag_coloridx_edited			4
#define tag_coloridx_edited_str		"04"
#define tag_coloridx_modified			5
#define tag_coloridx_modified_str		"05"

#define _CHAR(ofs) HT_EDIT_CHAR_CH ofs"00"tag_coloridx_unselected_str
#define _BYTE(ofs) HT_EDIT_BYTE_CH ofs"00"tag_coloridx_unselected_str
#define _WORD(ofs) HT_EDIT_WORD_CH ofs"0000"tag_coloridx_unselected_str
#define _DWORD(ofs) HT_EDIT_DWORD_CH ofs"00000000"tag_coloridx_unselected_str
#define _TIME(ofs) HT_EDIT_TIME_CH ofs"00000000"tag_coloridx_unselected_str
#define _SEL(unique_id, len) HT_SEL_CH unique_id len tag_coloridx_unselected_str"00000000"
#define _REF(unique_id, len, ref_id) HT_SEL_CH unique_id len tag_coloridx_unselected_str ref_id

int hexb(char *s);
int hexw(char *s);
int hexd(char *s);

char *findtag(char *htstring);
int tag_get_len(char *htstring);
int tag_get_vlen(char *htstring);
int tag_get_size(char *htstring);
dword tag_get_offset(char *htstring);
int tag_get_value(char *htstring);
void tag_set_color(char *htstring, int color);
void tag_set_offset(char *htstring, dword offset);
void tag_set_value(char *htstring, int value);
void tag_set_desc(char *htstring, void *desc);
int tag_is_editable(char *htstring);

int htstrlen(char *htstring);
int count_editables_in_group(char *htstring, int group);
int count_editables(char *htstring);
int count_groups(char *htstring);
char *get_editable(char *htstring, int n, int group);
char *get_group(char *htstring, int group);

char *htstrdup(char *string);

#endif /* !__HTSTRING_H__ */
