/* 
 *	The HTE Editor
 *	htunasm.cc
 *
 *	Copyright (C) 1999,2000 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "htopcode.h"
#include "htunasm.h"
#include "httools.h"

#include <stdarg.h>
#include <stdio.h>
#include <string.h>

static char *reg_names[3][8] = {
	{"al" ,"cl" ,"dl" ,"bl" ,"ah" ,"ch" ,"dh" ,"bh" },
	{"ax" ,"cx" ,"dx" ,"bx" ,"sp" ,"bp" ,"si" ,"di" },
	{"eax","ecx","edx","ebx","esp","ebp","esi","edi"}
};

static char *i_str[] =
	{"+eax","+ecx","+edx","+ebx","","+ebp","+esi","+edi"};

static char *sib_str[] =
	{"%p:[eax", "%p:[ecx", "%p:[edx", "%p:[ebx",
	 "%p:[esp", 0,	"%p:[esi", "%p:[edi" };

static char *r_str[] =
	{"bx+si","bx+di","bp+si","bp+di","si","di","bp","bx"};

/*
 *	CLASS ht_unasm
 */

#define mod(a)		(((a)>>6)&7)
#define reg(a)		(((a)>>3)&7)
#define rm(a)		((a)&7)

#define ss(a)		(((a)>>6)&7)
#define index(a)	(((a)>>3)&7)
#define base(a)	((a)&7)

ht_unasm::ht_unasm()
{
	opsize=addrsize=32;
}

ht_unasm::ht_unasm(int seg_size)
{
	set_segsize(seg_size);
}

ht_unasm::~ht_unasm()
{
}

void ht_unasm::set_segsize(int seg_size)
{
	this->stdopsize=this->stdaddrsize=seg_size;
}

STUB char* ht_unasm::addr_sym_func(ADDR addr)
{
	return 0;
}

unsigned char ht_unasm::getbyte()
{
	vaddr++;
     return *(bufp++);
}

unsigned short ht_unasm::getword()
{
     return *(bufp++)+(*(bufp++)<<8);
}

unsigned long ht_unasm::getdword()
{
     return *(bufp++)+(*(bufp++)<<8)+(*(bufp++)<<16)+(*(bufp++)<<24);
}

int ht_unasm::modrm()
{
	if (modrmv==-1) modrmv=getbyte();
     return modrmv;
}

int ht_unasm::sib()
{
	if (sibv==-1) sibv=getbyte();
     return sibv;
}

int ht_unasm::uprintf(char *format, ...)
{
	va_list args;
     va_start(args, format);
     int ret=vuprintf(format, args);
     va_end(args);
     return ret;
}

int ht_unasm::vuprintf(char *format, va_list args)
{
     int l=vsprintf(ubufp, format, args);
     ubufp+=l;
     return l;
}

void ht_unasm::uputc(char c)
{
	if (c == '\t')	{
		do {
          	uputc(' ');
		} while ((ubufp-ubuf) % 8);
	} else *ubufp++=c;
	*ubufp = 0;
}

int ht_unasm::bytesize(char c)
{
	switch (c) {
		case	'b':
			return 1;
		case	'w':
			return 2;
		case	'd':
			return 4;
		case	'v':
			if (opsize == 32)
				return 4;
			else
				return 2;
		case	'q':
			return 8;
	}
	return 0;
}

void ht_unasm::ua_ohex(char c, int extend, int defsize, int sign)
{
	char	fmt[8];
	int n=0, s=0, i;
	int delta;
	unsigned char buf[6];
	switch (c) {
		case	'a':
	 		break;
    		case	'b':
	 		n = 1;
	 		break;
    		case	'w':
	 		n = 2;
	 		break;
    		case	'd':
	 		n = 4;
	 		break;
    		case	's':
	 		n = 6;
	 		break;
    		case	'c':
    		case	'v':
	 		if (defsize ==	32) n = 4; else n = 2;
			break;
    		case	'p':
	 		if (defsize ==	32) n = 6; else n = 4;
	 		s = 1;
	 		break;
    		case	'x':
	 		return;
	}

     sprintf(fmt, "%%0%dx", extend ? extend*2 : n*2);

	for (i=0;	i<n;	i++)	buf[i] = getbyte();
	for (; i<extend; i++) buf[i] = (buf[i-1] & 0x80) ? 0xff : 0;
	if (s) {
		uprintf("%02x%02x:", buf[n-1], buf[n-2]);
    		n -=	2;
  	}
  	switch (n) {
    		case	1:
	 		delta = *(signed char *)buf;
	 		break;
   		case	2:
	 		delta = *(signed short *)buf;
	 		break;
    		case	4:
	 		delta = *(signed long *)buf;
			break;
	}
	if (sign >= 2)
		if (delta	< 0)	{
			delta = -delta;
			uputc('-');
		} else {
			uputc('+');
		}
	if (extend > n) {
		uprintf(fmt, delta);
		return;
	}
	if ((n ==	4) && sign < 2) {
		char *name=0;
		if ((name=addr_sym_func(delta))) {
			uprintf(name);
			return;
		}
	}
	switch (n) {
		case	1:
	 		uprintf(fmt, (unsigned char)delta);
	 		break;
    		case	2:
	 		uprintf(fmt, (unsigned short)delta);
	 		break;
    		case	4:
	 		uprintf(fmt, (unsigned long)delta);
	 		break;
	}
}

void ht_unasm::ua_reg_name(int which, char size)
{
	if (size == 'F') {
		uprintf("st(%d)", which);
		return;
	}
	if (((size == 'v') && (opsize	== 32)) || (size ==	'd')) {
    		uputc('e');
	}
	if (size=='q')	{
    		uprintf("mm%d", which);
  	} else if	(size == 'b') {
    		uputc("acdbacdb"[which]);
    		uputc("llllhhhh"[which]);
 	} else {
    		uputc("acdbsbsd"[which]);
    		uputc("xxxxppii"[which]);
  	}
}

int ht_unasm::ua_do_sib(int m)
{
	int pick_signed = default_pick_sign;
	int s, i,	b, extra=0;
  	s = ss(sib());
  	i = index(sib());
  	b = base(sib());
  	if (b == 5) {
		if (m == 0) {
			ua_str("%p:[");
			ua_ohex('d',	4, addrsize, 1);
		} else {
			ua_str("%p:[ebp");
			pick_signed |=	2;
		}
	} else {
		pick_signed |=	2;
		ua_str(sib_str[b]);
		if ((b ==	i) && (b != 4)	&& (i != 5)) extra = 1;
  	}
  	if (extra	== 0) {
		pick_signed |=	2;
		uprintf(i_str[i]);
	}
	if (i != 4 && s) uprintf("*%d", (1<<s)+extra);
	return pick_signed;
}

void ht_unasm::ua_do_modrm (char c)
{
	int m = mod(modrm());
	int r = rm(modrm());
	int extend = (addrsize == 32)	? 4 : 2;
	int pick_signed = default_pick_sign;
	if (m == 3) {
		ua_reg_name(r, c);
		return;
	}
	if ((m == 0) && (r == 5) && (addrsize == 32)) {
		ua_str("%p:[");
		ua_ohex('d',	extend, addrsize, 0);
		uputc(']');
		return;
	}
	if ((m == 0) && (r == 6) && (addrsize == 16)) {
		ua_str("%p:[");
		ua_ohex('w', extend, addrsize, 0);
		uputc(']');
		return;
	}
	if ((addrsize != 32) || (r != 4)) ua_str("%p:[");
	if (addrsize == 16) {
		uprintf(r_str[r]);
		pick_signed |= 2;
	} else {
		if (r == 4) pick_signed |= ua_do_sib(m); else {
			uprintf(reg_names[2][r]);
			pick_signed |= 2;
		}
	}
	ua_ohex("xbv"[m], extend, addrsize, pick_signed);
	uputc(']');
}

void ht_unasm::ua_floating_point(int e1)
{
	int esc = e1*8 + reg(modrm());
	if (mod(modrm()) == 3) {
		if (fspecial[esc]) {
			if (fspecial[esc][0] && (fspecial[esc][0][0] == '*')) {
				ua_str(fspecial[esc][0]+1);
			} else {
				ua_str(fspecial[esc][rm(modrm())]);
			}
		} else {
			ua_str(floatops[esc]);
			ua_str(" %EF");
		}
	} else {
		ua_str(floatops[esc]);
		ua_str(" %EF");
	}
}

void ht_unasm::ua_percent_str(char c, char **tptr)
{
	unsigned long vofs = 0, delta;
	int default_signed = default_pick_sign;
	char t = *(*tptr)++, it;
	int extend =	(addrsize	== 32) ? 4 : 2;
	int iextend;

	if (c != '+') {
		if (t == '-') {
			default_signed = 1;
			t =	*(*tptr)++;
		} else if (t == '+') {
			default_signed = 2;
			t =	*(*tptr)++;
		}
	}
	switch (c) {
		case 'A':
			ua_ohex(t, extend, addrsize, 0);
			break;
		case 'C':
			uprintf("cr%d", reg(modrm()));
			break;
		case 'D':
			uprintf("dr%d", reg(modrm()));
			break;
		case 'E':
			ua_do_modrm(t);
			break;
		case 'G':
			if (t ==	'F') ua_reg_name(rm(modrm()), t); else ua_reg_name(reg(modrm()), t);
			break;
		case 'I':
			it = *(*tptr)++;
			switch (t) {
				case 'b':
					iextend =	1;
					break;
				case 'v':
					iextend = extend;
					break;
				default:
					iextend =	0;
					break;
			}
			ua_ohex(it, iextend, opsize, default_signed);
			break;
		case 'J': {
			switch (bytesize(t)) {
				case 1:
					vofs	= (char)getbyte();
					break;
				case 2:
					vofs	= (unsigned char)getbyte();
					vofs	|= (unsigned char)getbyte()<<8;
					vofs=(short)vofs;
					break;
				case 4:
					vofs	= getbyte();
					vofs	|= getbyte() << 8;
					vofs	|= getbyte() << 16;
					vofs	|= getbyte() << 24;
					break;
			}
			char *name=addr_sym_func(vaddr+vofs);
			if (name) {
				uprintf("%s", name);
			} else if (delta) {
				uprintf("0x%08x (%c)", vaddr+vofs, (vofs & 0x80000000L) ? 0x18 :	0x19);
			}
			break;
		}
		case 'M':
			ua_do_modrm(t);
			break;
		case 'O':
			ua_str("%p:[");
			ua_ohex(t, extend, addrsize, 0);
			uputc(']');
			break;
		case 'P':
			uprintf("mm%d", reg(modrm()));
			break;
		case 'Q':
			if (mod(modrm())==3) {
				ua_reg_name(rm(modrm()), 'q');
			} else {
				ua_do_modrm(t);
			}
			break;
		case 'R':
			ua_do_modrm(t);
			break;
		case 'S':
			uputc("ecsdfg"[reg(modrm())]);
			uputc('s');
			break;
		case 'T':
			uprintf("tr%d", reg(modrm()));
			break;
		case 'X':
			uprintf("ds:[");
			if (addrsize == 32) uputc('e');
			uprintf("si]");
			break;
		case 'Y':
			uprintf("es:[");
			if (addrsize == 32) uputc('e');
			uprintf("di]");
			break;
		case '2':
			ua_str(opmap_ext[getbyte()]);
			break;
		case '3': {
			ua_str("%Pq, %Qq");
			char *n=opmap_ext2[getbyte()];
			if (n) {
				char	b[200];
				memmove(b, ubuf, ubufp-ubuf+1);
				ubufp=ubuf;
				uprintf(n);
				uputc('\t');
				uprintf((char*)b);
			} else {
				invalid=1;
		 	}
		 	break;
		}
		case 'e':
			if (opsize ==	32) {
				if (t == 'w') uputc('d'); else {
					uputc('e');
					uputc(t);
				}
			} else uputc(t);
			break;
		case 'f':
			ua_floating_point(t-'0');
			break;
		case 'g':
			if (t>'9') ua_str(groups[t-'A'+10][reg(modrm())]); else
				ua_str(groups[t-'0'][reg(modrm())]);
			break;
		case 'p':
			switch (t) {
				case 'c':
				case 'd':
				case 'e':
				case 'f':
				case 'g':
				case 's':
					prefix = t;
					ua_str(opmap[getbyte()]);
					break;
				case ':':
					if (prefix) uprintf("%cs:", prefix);
					break;
				case ' ':
					ua_str(opmap[getbyte()]);
					break;
			}
			break;
		case 's':
			switch (t) {
				case 'a':
					addrsize = 48 - stdaddrsize;
					ua_str(opmap[getbyte()]);
					break;
				case 'o':
					opsize = 48 - stdopsize;
					ua_str(opmap[getbyte()]);
					break;
			}
			break;
		case '+':
			switch (t) {
				case '-':
					default_pick_sign = 1;
					break;
				case '+':
					default_pick_sign =	2;
					break;
				default:
					default_pick_sign = 0;
					break;
			}
               break;
	}
}

void ht_unasm::ua_str(char *s)
{
	int c;
	if (!s) {
		invalid=1;
		return;
	}
	while ((c = *s++) != 0) {
		if (c == '%') {
			c = *s++;
			ua_percent_str(c, &s);
		} else uputc(c);
	}
}

void ht_unasm::size_ohex(char c, int defsize)
{
	int n=0;
	switch (c) {
		case	'a':
	 		break;
    		case	'b':
	 		n = 1;
	 		break;
    		case	'w':
	 		n = 2;
	 		break;
    		case	'd':
	 		n = 4;
	 		break;
    		case	's':
	 		n = 6;
	 		break;
    		case	'c':
    		case	'v':
	 		if (defsize ==	32) n = 4; else n = 2;
			break;
    		case	'p':
	 		if (defsize ==	32) n = 6; else n = 4;
	 		break;
    		case	'x':
	 		return;
	}
     bufp+=n;
}

void ht_unasm::size_do_sib(int m)
{
	int b;
  	b = base(sib());
  	if (b == 5) {
		if (m == 0) {
          	bufp+=4;
		}
	} else {
		size_str(sib_str[b]);
  	}
}

void ht_unasm::size_do_modrm (char c)
{
	int m = mod(modrm());
	int r = rm(modrm());
     if (m == 3) return;
	if ((m == 0) && (r == 5) && (addrsize == 32)) {
     	bufp+=4;
		return;
	}
	if ((m == 0) && (r == 6) && (addrsize == 16)) {
     	bufp+=2;
		return;
	}
	if ((addrsize == 32) && (r == 4)) size_do_sib(m);
     size_ohex("xbv"[m], addrsize);
}

void ht_unasm::size_floating_point(int e1)
{
	int esc = e1*8 + reg(modrm());
	if (mod(modrm()) == 3) {
		if (fspecial[esc]) {
			if (fspecial[esc][0] && (fspecial[esc][0][0] == '*')) {
				size_str(fspecial[esc][0]+1);
			} else {
				size_str(fspecial[esc][rm(modrm())]);
			}
		} else {
			size_str(floatops[esc]);
			size_str(" %EF");
		}
	} else {
		size_str(floatops[esc]);
		size_str(" %EF");
	}
}

void ht_unasm::size_percent_str(char c, char **tptr)
{
	char t = *(*tptr)++;
	if (c != '+') {
		if (t == '-') {
			t =	*(*tptr)++;
		} else if (t == '+') {
			t =	*(*tptr)++;
		}
	}
	switch (c) {
		case 'A':
			size_ohex(t, addrsize);
			break;
		case 'C':
          	modrm();
			break;
		case 'D':
          	modrm();
			break;
		case 'E':
			size_do_modrm(t);
			break;
		case 'G':
			modrm();
			break;
		case 'I':
			size_ohex(*(*tptr)++, opsize);
			break;
		case 'J': {
			bufp+=bytesize(t);
			break;
		}
		case 'M':
			size_do_modrm(t);
			break;
		case 'O':
			size_ohex(t, addrsize);
			break;
		case 'P':
			modrm();
			break;
		case 'Q':
			if (mod(modrm())!=3) {
				size_do_modrm(t);
			}
			break;
		case 'R':
			size_do_modrm(t);
			break;
		case 'S':
			modrm();
			break;
		case 'T':
			modrm();
			break;
		case 'X':
			break;
		case 'Y':
			break;
		case '2':
			size_str(opmap_ext[getbyte()]);
			break;
		case '3': {
	 		size_str("%Pq, %Qq");
               getbyte();
		 	break;
		}
		case 'e':
			break;
		case 'f':
			size_floating_point(t-'0');
			break;
		case 'g':
			if (t>'9') size_str(groups[t-'A'+10][reg(modrm())]); else
				size_str(groups[t-'0'][reg(modrm())]);
			break;
		case 'p':
			switch (t) {
				case 'c':
				case 'd':
				case 'e':
				case 'f':
				case 'g':
				case 's':
					prefix = t;
					size_str(opmap[getbyte()]);
					break;
				case ':':
					break;
				case ' ':
					size_str(opmap[getbyte()]);
					break;
			}
			break;
		case 's':
			switch (t) {
				case 'a':
					addrsize = 48 - stdaddrsize;
					size_str(opmap[getbyte()]);
					break;
				case 'o':
					opsize = 48 - stdopsize;
					size_str(opmap[getbyte()]);
					break;
			}
			break;
		case '+':
               break;
	}
}

void ht_unasm::size_str(char *s)
{
	int c;
	if (!s) {
		invalid=1;
		return;
	}
	while ((c = *s++) != 0) {
		if (c == '%') {
			c = *s++;
			size_percent_str(c, &s);
		};
	}
}

char *ht_unasm::getstring(unsigned char *code, int maxsize, int addr, int *size)
{
	bufp=code;
     ubufp=ubuf;
	default_pick_sign=0;
     vaddr=addr;
     invalid=0;
     modrmv=sibv=-1;
     prefix=0;
     opsize=stdopsize;
     addrsize=stdaddrsize;
     ua_str(opmap[getbyte()]);
     if ((invalid) || (bufp-code>MIN(maxsize, 15))) {
     	*size=1;
     	return "<invalid>";
	} else {
     	*size=bufp-code;
		return ubuf;
	}
}

int ht_unasm::getsize(unsigned char *code, int maxsize)
{
	bufp=code;
     ubufp=ubuf;
	default_pick_sign=0;
     vaddr=0;
     invalid=0;
     modrmv=sibv=-1;
     prefix=0;
     opsize=stdopsize;
     addrsize=stdaddrsize;
	size_str(opmap[getbyte()]);
     if ((invalid) || (bufp-code>MIN(maxsize, 15))) {
     	return 1;
	} else {
     	return bufp-code;
	}
}

