/*
 *   htcurses.cc - DJGPP implementation
 */

#include "global.h"
#include "htcurses.h"
#include "htlayout.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <conio.h>
#include <dpmi.h>
#include <go32.h>
#include <pc.h>

/*
 *	CLASS screendrawbuf
 */

screendrawbuf::screendrawbuf()
{
	bounds b;
     buf=0;
     cursorx=0;
     cursory=0;

     b.x=0;
     b.y=0;
	b.w=ScreenCols();
     b.h=ScreenRows();
     setbounds(&b);

	screensel=__dpmi_allocate_ldt_descriptors(1);
	__dpmi_set_descriptor_access_rights(screensel, 0xc0f3);
	__dpmi_set_segment_base_address(screensel, 0xb8000);
	__dpmi_set_segment_limit(screensel, 0x7fff);

     show();
}

screendrawbuf::~screendrawbuf()
{
	if (buf) {
          delete buf;
     }
}

int screendrawbuf::print(int x, int y, int c, char *text)
{
     return lprint(x, y, c, 0x7fffffff, text);
}

int screendrawbuf::lprint(int x, int y, int c, int l, char *text)
{
	int n=0;
	unsigned short *b=buf+x+y*size.w;
	while ((*text) && (n<l)) {
		*(b++)=(c<<8)|(unsigned char)*(text++);
          n++;
	}
     return n;
}

void screendrawbuf::fill(int x, int y, int w, int h, int c)
{
     for (int i=0; i<h; i++) {
		unsigned short *b=buf+x+(y+i)*size.w;
     	for (int j=0; j<w; j++) {
			*(b++)=(c<<8)|' ';
     	}
     }
}

void screendrawbuf::setbounds(bounds *b)
{
	genericdrawbuf::setbounds(b);
	if (buf) delete buf;
	buf=(unsigned short *)malloc(sizeof *buf * size.w * size.h);
     memset(buf, 0, sizeof *buf * size.w * size.h);
}

void screendrawbuf::drawbuffer(drawbuf *b, int x, int y)
{
	drawbufch *ch=b->buf;
	for (int iy=0; iy<b->size.h; iy++) {
		unsigned short *k=buf+x+(iy+y)*size.w;
		for (int ix=0; ix<b->size.w; ix++) {
			*(k++)=(ch->c<<8)|((byte)ch->ch);
			ch++;
		}
     }
}

void screendrawbuf::show()
{
	gotoxy(cursorx+1, cursory+1);
//	waitretrace();           /* FIXME: wait or not ??? */
	movedata(_go32_my_ds(), (int)buf, screensel, 0, size.w*size.h*2);
}

void screendrawbuf::getcursor(int *x, int *y)
{
	*x=cursorx;
     *y=cursory;
}

void screendrawbuf::hidecursor()
{
	setcursor(0, 0);
}

void screendrawbuf::setcursor(int x, int y)
{
	cursorx=x;
	cursory=y;
}

void screendrawbuf::showcursor()
{
}


void waitretrace()
{
	asm ("
          mov	$0x3da, %edx
waitretrace_eot:
          in	%dx, %al
          test	$8, %al
          jnz	waitretrace_eot
waitretrace_eor:
          in	%dx, %al
          test $8, %al
          jz	waitretrace_eor");
}

/*
 *	CLASS drawbuf
 */

drawbuf::drawbuf(bounds *b)
{
     buf=0;
     setbounds(b);
}

drawbuf::~drawbuf()
{
	if (buf) {
          delete buf;
     }
}

int drawbuf::print(int x, int y, int c, char *text)
{
     return lprint(x, y, c, 0x7fffffff, text);
}

int drawbuf::lprint(int x, int y, int c, int l, char *text)
{
	int ox=0;
     x-=size.x;
     y-=size.y;
	drawbufch *b=buf+x+y*size.w;
     if (y<size.h) {
	     while ((ox<size.w) && (*text) && (ox<l)) {
     	     b->ch=(byte)*text;
          	b->c=c;
	     	text++;
     		ox++;
               b++;
	     }
	}
     return ox;
}

void drawbuf::fill(int x, int y, int w, int h, int c)
{
	x-=size.x;
     y-=size.y;
	drawbufch *b=buf+x+y*size.w;
	for (int iy=y; iy<y+h; iy++) {
		for (int ix=x; ix<x+w; ix++) {
			b->ch=' ';
			b->c=c;
			b++;
		}
     }
}

void drawbuf::setbounds(bounds *b)
{
	genericdrawbuf::setbounds(b);
	if (buf) delete buf;
	buf=(drawbufch*)malloc(sizeof *buf * size.w * size.h);
     memset(buf, 0, sizeof *buf * size.w * size.h);
}

/*****************************************************************************
 ****************************************************************************/

genericdrawbuf::genericdrawbuf()
{
}

genericdrawbuf::~genericdrawbuf()
{
}

void genericdrawbuf::fill(int x, int y, int w, int h, int c)
{
}

int genericdrawbuf::printf(int x, int y, int c, char *format, ...)
{
	char buf[1024];
	va_list arg;
     va_start(arg, format);
	vsprintf(buf, format, arg);
     va_end(arg);
     return print(x, y, c, buf);
}

int genericdrawbuf::print(int x, int y, int c, char *text)
{
	return 0;
}

int genericdrawbuf::lprint(int x, int y, int c, int l, char *text)
{
	return 0;
}

void genericdrawbuf::setbounds(bounds *b)
{
	size=*b;
}


